/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse

    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

/***************************************************************************

  DIKeyboard.c

 ***************************************************************************/

#include "driver.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#define DIRECTINPUT_VERSION 0x0700
#include <dinput.h>
#include "MAME32.h"
#include "M32Util.h"
#include "DirectInput.h"
#include "Keyboard.h"
#include "DIKeyboard.h"
#include "DIJoystick.h"
#include "status.h"
#include "dxdecode.h"
#include "ui_text.h"
#include "win32text.h"

/***************************************************************************
    function prototypes
 ***************************************************************************/

static int          DIKeyboard_init(options_type *options);
static void         DIKeyboard_exit(void);

static const struct KeyboardInfo * DIKeyboard_get_key_list(void);
static void         DIKeyboard_customize_inputport_defaults(struct ipd *defaults);
static int          DIKeyboard_is_key_pressed(int keycode);
static int          DIKeyboard_readkey_unicode(int flush);

static BOOL         DIKeyboard_OnMessage(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam, LRESULT* pResult);
static void         OnActivateApp(HWND hWnd, BOOL fActivate, DWORD dwThreadId);
static void         OnKey(HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags);
static void         OnChar(HWND hWnd, TCHAR ch, int cRepeat);

static void         DIKeyboard_PollKeyboard(void);

/***************************************************************************
    External variables
 ***************************************************************************/

struct OSDKeyboard  DIKeyboard =
{
    { DIKeyboard_init },                         /* init                         */
    { DIKeyboard_exit },                         /* exit                         */
    { DIKeyboard_get_key_list },                 /* get_key_list                 */
    { DIKeyboard_customize_inputport_defaults }, /* customize_inputport_defaults */
    { DIKeyboard_is_key_pressed },               /* is_key_pressed               */
    { DIKeyboard_readkey_unicode },              /* readkey_unicode              */

    { DIKeyboard_OnMessage },                    /* OnMessage                    */
};

/***************************************************************************
    Internal structures
 ***************************************************************************/

struct tKeyboard_private
{
    LPDIRECTINPUTDEVICE7 m_didKeyboard;
    byte                 m_key[256];
    int                  m_DefaultInput;
    TCHAR                m_chPressed;
};

/***************************************************************************
    Internal variables
 ***************************************************************************/

static struct tKeyboard_private This;

/***************************************************************************
    External OSD functions
 ***************************************************************************/

/*
    put here anything you need to do when the program is started. Return 0 if
    initialization was successful, nonzero otherwise.
*/
static int DIKeyboard_init(options_type *options)
{
    HRESULT hResult;
    int i;

    This.m_didKeyboard  = NULL;
    This.m_DefaultInput = options->default_input;
    This.m_chPressed    = 0;

    for (i = 0; i < 256; i++)
       This.m_key[i] = 0;


    if (di == NULL)
    {
        ErrorMsg(win32_getstring(WIN32_dinput_notinitialized));
        return 1;
    }

    /* L[{[hp̃foCXIuWFNg쐬 */
    hResult = IDirectInput7_CreateDeviceEx(di,
                                           &GUID_SysKeyboard,
                                           &IID_IDirectInputDevice7,
                                           &This.m_didKeyboard,
                                           NULL);
    if (FAILED(hResult))
    {
        ErrorMsg("IDirectInput7.CreateDeviceEx(keyboard) %s%s",
				 win32_getstring(WIN32_errorcode),
                 DirectXDecodeError(hResult));
        return 1;
    }

    /* L[{[hp̃f[^tH[}bgݒ */
    hResult = IDirectInputDevice7_SetDataFormat(This.m_didKeyboard,
                                                &c_dfDIKeyboard);
    if (FAILED(hResult))
    {
        ErrorMsg("IDirectInput7.SetDataFormat(keyboard) %s%s",
				 win32_getstring(WIN32_errorcode),
                 DirectXDecodeError(hResult));
        return 1;
    }

    /* xݒ (tHAOEhr[h) */
    hResult = IDirectInputDevice7_SetCooperativeLevel(This.m_didKeyboard,
                                                      MAME32App.m_hWnd,
                                                      DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
    if (FAILED(hResult))
    {
        ErrorMsg("IDirectInput7.SetCooperativeLevel(keyboard) %s%s",
				 win32_getstring(WIN32_errorcode),
                 DirectXDecodeError(hResult));
        return 1;
    }

    /* ͐Jn */
    hResult = IDirectInputDevice7_Acquire(This.m_didKeyboard);

    return 0;
}

/*
  put here cleanup routines to be executed when the program is terminated.
  */
static void DIKeyboard_exit(void)
{
    if (!This.m_didKeyboard)
        return;

    /*
     Cleanliness is next to godliness.  Unacquire the device
     one last time just in case we got really confused and tried
     to exit while the device is still acquired.
     */
    IDirectInputDevice7_Unacquire(This.m_didKeyboard);
    IDirectInputDevice7_Release(This.m_didKeyboard);

    This.m_didKeyboard = NULL;
}

static struct KeyboardInfo keylist[] =
{
	{ keyname(akey),			DIK_A,				KEYCODE_A },
	{ keyname(bkey),			DIK_B,				KEYCODE_B },
	{ keyname(ckey),			DIK_C,				KEYCODE_C },
	{ keyname(dkey),			DIK_D,				KEYCODE_D },
	{ keyname(ekey),			DIK_E,				KEYCODE_E },
	{ keyname(fkey),			DIK_F,				KEYCODE_F },
	{ keyname(gkey),			DIK_G,				KEYCODE_G },
	{ keyname(hkey),			DIK_H,				KEYCODE_H },
	{ keyname(ikey),			DIK_I,				KEYCODE_I },
	{ keyname(jkey),			DIK_J,				KEYCODE_J },
	{ keyname(kkey),			DIK_K,				KEYCODE_K },
	{ keyname(lkey),			DIK_L,				KEYCODE_L },
	{ keyname(mkey),			DIK_M,				KEYCODE_M },
	{ keyname(nkey),			DIK_N,				KEYCODE_N },
	{ keyname(okey),			DIK_O,				KEYCODE_O },
	{ keyname(pkey),			DIK_P,				KEYCODE_P },
	{ keyname(qkey),			DIK_Q,				KEYCODE_Q },
	{ keyname(rkey),			DIK_R,				KEYCODE_R },
	{ keyname(skey),			DIK_S,				KEYCODE_S },
	{ keyname(tkey),			DIK_T,				KEYCODE_T },
	{ keyname(ukey),			DIK_U,				KEYCODE_U },
	{ keyname(vkey),			DIK_V,				KEYCODE_V },
	{ keyname(wkey),			DIK_W,				KEYCODE_W },
	{ keyname(xkey),			DIK_X,				KEYCODE_X },
	{ keyname(ykey),			DIK_Y,				KEYCODE_Y },
	{ keyname(zkey),			DIK_Z,				KEYCODE_Z },
	{ keyname(0key),			DIK_0,				KEYCODE_0 },
	{ keyname(1key),			DIK_1,				KEYCODE_1 },
	{ keyname(2key),			DIK_2,				KEYCODE_2 },
	{ keyname(3key),			DIK_3,				KEYCODE_3 },
	{ keyname(4key),			DIK_4,				KEYCODE_4 },
	{ keyname(5key),			DIK_5,				KEYCODE_5 },
	{ keyname(6key),			DIK_6,				KEYCODE_6 },
	{ keyname(7key),			DIK_7,				KEYCODE_7 },
	{ keyname(8key),			DIK_8,				KEYCODE_8 },
	{ keyname(9key),			DIK_9,				KEYCODE_9 },
	{ keyname(0pad),			DIK_NUMPAD0,		KEYCODE_0_PAD },
	{ keyname(1pad),			DIK_NUMPAD1,		KEYCODE_1_PAD },
	{ keyname(2pad),			DIK_NUMPAD2,		KEYCODE_2_PAD },
	{ keyname(3pad),			DIK_NUMPAD3,		KEYCODE_3_PAD },
	{ keyname(4pad),			DIK_NUMPAD4,		KEYCODE_4_PAD },
	{ keyname(5pad),			DIK_NUMPAD5,		KEYCODE_5_PAD },
	{ keyname(6pad),			DIK_NUMPAD6,		KEYCODE_6_PAD },
	{ keyname(7pad),			DIK_NUMPAD7,		KEYCODE_7_PAD },
	{ keyname(8pad),			DIK_NUMPAD8,		KEYCODE_8_PAD },
	{ keyname(9pad),			DIK_NUMPAD9,		KEYCODE_9_PAD },
	{ keyname(function1),		DIK_F1,				KEYCODE_F1 },
	{ keyname(function2),		DIK_F2,				KEYCODE_F2 },
	{ keyname(function3),		DIK_F3,				KEYCODE_F3 },
	{ keyname(function4),		DIK_F4,				KEYCODE_F4 },
	{ keyname(function5),		DIK_F5,				KEYCODE_F5 },
	{ keyname(function6),		DIK_F6,				KEYCODE_F6 },
	{ keyname(function7),		DIK_F7,				KEYCODE_F7 },
	{ keyname(function8),		DIK_F8,				KEYCODE_F8 },
	{ keyname(function9),		DIK_F9,				KEYCODE_F9 },
	{ keyname(function10),		DIK_F10,			KEYCODE_F10 },
	{ keyname(function11),		DIK_F11,			KEYCODE_F11 },
	{ keyname(function12),		DIK_F12,			KEYCODE_F12 },
	{ keyname(escape),			DIK_ESCAPE,			KEYCODE_ESC },
	{ keyname(tilde),			DIK_GRAVE,			KEYCODE_TILDE },
	{ keyname(minus),			DIK_MINUS,			KEYCODE_MINUS },
	{ keyname(equals),			DIK_EQUALS,			KEYCODE_EQUALS },
	{ keyname(backspace),		DIK_BACKSPACE,		KEYCODE_BACKSPACE },
	{ keyname(tab),				DIK_TAB,			KEYCODE_TAB },
	{ keyname(openbrace),		DIK_LBRACKET,		KEYCODE_OPENBRACE },
	{ keyname(closebrace),		DIK_RBRACKET,		KEYCODE_CLOSEBRACE },
	{ keyname(enter),			DIK_RETURN,			KEYCODE_ENTER },
	{ keyname(colon),			DIK_COLON,			KEYCODE_COLON },
	{ keyname(quote),			DIK_APOSTROPHE,		KEYCODE_QUOTE },
	{ keyname(backslash),		DIK_BACKSLASH,		KEYCODE_BACKSLASH },
	{ keyname(comma),			DIK_COMMA,			KEYCODE_COMMA },
	{ keyname(stop),			DIK_PERIOD,			KEYCODE_STOP },
	{ keyname(slash),			DIK_SLASH,			KEYCODE_SLASH },
	{ keyname(space),			DIK_SPACE,			KEYCODE_SPACE },
	{ keyname(insert),			DIK_INSERT,			KEYCODE_INSERT },
	{ keyname(del),				DIK_DELETE,			KEYCODE_DEL },
	{ keyname(home),			DIK_HOME,			KEYCODE_HOME },
	{ keyname(end),				DIK_END,			KEYCODE_END },
	{ keyname(pageup),			DIK_PGUP,			KEYCODE_PGUP },
	{ keyname(pagedown),		DIK_PGDN,			KEYCODE_PGDN },
	{ keyname(left),			DIK_LEFT,			KEYCODE_LEFT },
	{ keyname(right),			DIK_RIGHT,			KEYCODE_RIGHT },
	{ keyname(up),				DIK_UP,				KEYCODE_UP },
	{ keyname(down),			DIK_DOWN,			KEYCODE_DOWN },
	{ keyname(slashpad),		DIK_DIVIDE, 		KEYCODE_SLASH_PAD },
	{ keyname(asterisk),		DIK_MULTIPLY,		KEYCODE_ASTERISK },
	{ keyname(minuspad),		DIK_SUBTRACT,		KEYCODE_MINUS_PAD },
	{ keyname(pluspad),			DIK_ADD,		    KEYCODE_PLUS_PAD },
	{ keyname(delpad),			DIK_DECIMAL,		KEYCODE_DEL_PAD },
	{ keyname(enterpad),		DIK_NUMPADENTER,	KEYCODE_ENTER_PAD },
//	{ keyname(printscreen),		DIK_PRTSCR,			KEYCODE_PRTSCR },
//	{ keyname(pause),			DIK_PAUSE,			KEYCODE_PAUSE },
	{ keyname(leftshift),		DIK_LSHIFT,			KEYCODE_LSHIFT },
	{ keyname(rightshift),		DIK_RSHIFT,			KEYCODE_RSHIFT },
	{ keyname(leftcontrol),		DIK_LCONTROL,		KEYCODE_LCONTROL },
	{ keyname(rightcontrol),	DIK_RCONTROL,		KEYCODE_RCONTROL },
	{ keyname(leftalt),			DIK_LALT,			KEYCODE_LALT },
	{ keyname(rightalt),		DIK_RALT,			KEYCODE_RALT },
	{ keyname(leftwin),			DIK_LWIN,			KEYCODE_OTHER },
	{ keyname(rightwin),		DIK_RWIN,			KEYCODE_OTHER },
	{ keyname(menu),			DIK_APPS,			KEYCODE_OTHER },
	{ keyname(scrolllock),		DIK_SCROLL,		    KEYCODE_SCRLOCK },
	{ keyname(numlock),			DIK_NUMLOCK,		KEYCODE_NUMLOCK },
	{ keyname(capslock),		DIK_CAPSLOCK,		KEYCODE_CAPSLOCK },

	{ 0, 0, 0 }	/* end of table */
};

/*
  return a list of all available keys (see input.h)
*/
static const struct KeyboardInfo * DIKeyboard_get_key_list(void)
{
    return keylist;
}

/*
  inptport.c defines some general purpose defaults for key bindings. They may be
  further adjusted by the OS dependant code to better match the available keyboard,
  e.g. one could map pause to the Pause key instead of P, or snapshot to PrtScr
  instead of F12. Of course the user can further change the settings to anything
  he/she likes.
  This function is called on startup, before reading the configuration from disk.
  Scan the list, and change the keys you want.
*/
static void DIKeyboard_customize_inputport_defaults(struct ipd *defaults)
{
    Keyboard_CustomizeInputportDefaults(This.m_DefaultInput, defaults);
}

/*
  tell whether the specified key is pressed or not. keycode is the OS dependant
  code specified in the list returned by osd_customize_inputport_defaults().
*/
static int DIKeyboard_is_key_pressed(int keycode)
{
    if (keycode == DIK_ESCAPE && MAME32App.Done())
    {
        return 1;
    }

    MAME32App.ProcessMessages();

    DIKeyboard_PollKeyboard();

    return (This.m_key[keycode] & 0x80) != 0;
}

static int DIKeyboard_readkey_unicode(int flush)
{
    if (flush)
    {
        This.m_chPressed = 0;
        return 0;
    }

    MAME32App.ProcessMessages();

    if (This.m_chPressed)
    {
        TCHAR ch = This.m_chPressed;

        This.m_chPressed = 0;
        return ch;
    }

    return 0;
}


BOOL DIKeyboard_Available(void)
{
    HRESULT hResult;
    GUID    guidNULL = {0, 0, 0, {0, 0, 0, 0, 0, 0, 0, 0}};
    GUID    guidDevice = guidNULL;

    if (di == NULL)
    {
        return FALSE;
    }

    /* enumerate for keyboard devices */
    hResult = IDirectInput7_EnumDevices(di,
                                        DIDEVTYPE_KEYBOARD,
                                        (LPDIENUMDEVICESCALLBACK)inputEnumDeviceProc,
                                        &guidDevice,
                                        DIEDFL_ATTACHEDONLY);
    if (FAILED(hResult))
    {
        return FALSE;
    }

    if (!IsEqualGUID(&guidDevice, &guidNULL))
    {
        return TRUE;
    }

    return FALSE;
}

/***************************************************************************
    Message handlers
 ***************************************************************************/

static BOOL DIKeyboard_OnMessage(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam, LRESULT* pResult)
{
    switch (Msg)
    {
        PEEK_MESSAGE(hWnd, WM_ACTIVATEAPP,  OnActivateApp);
        HANDLE_MESSAGE(hWnd, WM_KEYDOWN,    OnKey);
        HANDLE_MESSAGE(hWnd, WM_KEYUP,      OnKey);
        HANDLE_MESSAGE(hWnd, WM_CHAR,       OnChar);
    }
    return FALSE;
}

static void OnActivateApp(HWND hWnd, BOOL fActivate, DWORD dwThreadId)
{
    if (!This.m_didKeyboard)
        return;

    if (MAME32App.m_bIsActive == TRUE)
        IDirectInputDevice7_Acquire(This.m_didKeyboard);
    else
        IDirectInputDevice7_Unacquire(This.m_didKeyboard);

}

static void OnKey(HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags)
{
}

static void OnChar(HWND hWnd, TCHAR ch, int cRepeat)
{
    This.m_chPressed = ch;
}

/***************************************************************************
  Internal functions
 ***************************************************************************/

static void DIKeyboard_PollKeyboard(void)
{
    HRESULT hResult;

    memset(This.m_key, 0, 256 * sizeof(byte));

    if (This.m_didKeyboard == NULL)
       return;

again:

    hResult = IDirectInputDevice7_GetDeviceState(This.m_didKeyboard,sizeof(This.m_key),&This.m_key);
    if (hResult == DIERR_INPUTLOST || hResult == DIERR_NOTACQUIRED)
    {
        /*
            DirectInput is telling us that the input stream has
            been interrupted.  We aren't tracking any state
            between polls, so we don't have any special reset
            that needs to be done.  We just re-acquire and
            try again.
        */
        hResult = IDirectInputDevice7_Acquire(This.m_didKeyboard);
        if (SUCCEEDED(hResult))
            goto again;
    }
}
