/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse

    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

***************************************************************************/

/***************************************************************************

  Properties.c

    Properties Popup and Misc UI support routines.

    Created 8/29/98 by Mike Haaland (mhaaland@hypertech.com)

***************************************************************************/

#define MULTI_MONITOR

#include "driver.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <commctrl.h>
#include <commdlg.h>
#include <strings.h>
#include <sys/stat.h>
#include <ddraw.h>

#include "info.h"
#include "audit32.h"
#include "options.h"
#include "file.h"
#include "resource.h"
#include "Joystick.h"       /* For joystick avalibility. */
#include "DIJoystick.h"     /* For DIJoystick avalibility. */
#include "DIKeyboard.h"     /* For DIKeyboard avalibility. */
#include "DDrawDisplay.h"   /* For display modes. */
#include "m32util.h"
#include "directdraw.h"
#include "properties.h"

#include "Mame32.h"
#include "DataMap.h"
#include "resource.hm"
#include "win32text.h"

#include <multimon.h>

/*
   For compilers that don't support nameless unions
*/
#ifdef NONAMELESSUNION
/* PROPSHEETHEADER*/
#define nStartPage  DUMMYUNIONNAME2.nStartPage
/* PROPSHEETPAGE */
#define pszTemplate DUMMYUNIONNAME.pszTemplate
#define ppsp        DUMMYUNIONNAME3.ppsp
/* PROPSHEETHEADER */
#define pszIcon     DUMMYUNIONNAME.pszIcon
#endif

/***************************************************************
 * Imported function prototypes
 ***************************************************************/

extern BOOL GameUsesTrackball(int game);
extern BOOL GameUsesRasterDriver(int game);
extern BOOL GameUsesRomPatch(int game);
extern BOOL GameUsesOverClock(int game);
extern BOOL GameOnlyJukeBox(int game);
extern int load_driver_history(const struct GameDriver *drv, char *buffer, int bufsize);
extern int load_driver_command(const struct GameDriver *drv, char *buffer, int bufsize);

extern void neogeo_set_driver(int game);

/**************************************************************
 * Local function prototypes
 **************************************************************/

static INT_PTR CALLBACK GamePropertiesDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static INT_PTR CALLBACK GameDisplayOptionsProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);

static void SetStereoEnabled(HWND hWnd, int index);
static void InitializeOptions(HWND hDlg);
static void InitializeMisc(HWND hDlg);
static void OptOnHScroll(HWND hWnd, HWND hwndCtl, UINT code, int pos);
static void GammaSelectionChange(HWND hwnd);
static void BrightnessSelectionChange(HWND hwnd);
static void TransparentChange(HWND hwnd);
static void DisplayChange(HWND hWnd);
static void VolumeSelectionChange(HWND hwnd);
static void DSoundBufferSelectionChange(HWND hwnd);
static void RandomTimeChange(HWND hwnd);
static void UpdateDisplayModeUI(HWND hwnd, DWORD dwDepth);
static void InitializeDisplayModeUI(HWND hwnd);
static void InitializeSoundUI(HWND hwnd);
static void InitializeDisplayTypeUI(HWND hwnd);
static void InitializeRender(HWND hwnd);
static void InitializeSkippingUI(HWND hwnd);
static void InitializeDefaultInputUI(HWND hWnd);
static void InitializeDIjoyUI(HWND hwnd);
static void InitializeCoreUI(HWND hWnd);
static void InitializeDriverUI(HWND hWnd);
static void InitializeBiosUI(HWND hWnd);
static void PropToOptions(HWND hWnd, options_type *o);
static void OptionsToProp(HWND hWnd, options_type *o);
static void SetPropEnabledControls(HWND hWnd);

static void BuildDataMap(void);
static void ResetDataMap(void);

static void FixBuf(char *buf);

/**************************************************************
 * Local private variables
 **************************************************************/

static options_type origGameOpts, * lpGameOpts = NULL;

static int  iGame = 0;
static BOOL bUseDefaults = FALSE;
static BOOL bReset = FALSE;
static BOOL bFullScreen = 0;
static int  nSampleRate = 0;
static int  nVolume = 0;
static int  nScreenSize = 0;
static int  nGamma = 0;
static int  nDSBuffer = 0;
static int  nRandomTime = 0;

/* Game history variables */
#define MAX_DATAFILE_LEN     (65536)

static char historyBuf[MAX_DATAFILE_LEN];
static char commandBuf[MAX_DATAFILE_LEN];
static char tempConvertBuf[MAX_DATAFILE_LEN];

/* Property sheets */
static DWORD dwDlgId[] = {
    IDD_PROP_GAME,
    IDD_PROP_AUDIT,
    IDD_PROP_DISPLAY,
    IDD_PROP_SOUND,
    IDD_PROP_INPUT,
    IDD_PROP_DRIVER,
    IDD_PROP_MISC
};

#define NUM_PROPSHEETS (sizeof(dwDlgId) / sizeof(dwDlgId[0]))

/***************************************************************
 * Public functions
 ***************************************************************/

/* Checks of all ROMs are available for 'game' and returns result
 * Returns TRUE if all ROMs found, 0 if any ROMs are missing.
 */
BOOL FindRomSet(int game)
{
    const struct RomModule  *region, *rom;
    const struct GameDriver *gamedrv;
    const char              *name;
    int                     err;
    unsigned int            length, icrc;

    gamedrv = drivers[game];

    if (!osd_faccess(gamedrv->name, OSD_FILETYPE_ROM))
    {
        /* if the game is a clone, try loading the ROM from the main version */
        if (gamedrv->clone_of == 0
        ||  gamedrv->clone_of == DRIVER_ROOT
        ||  !osd_faccess(gamedrv->clone_of->name, OSD_FILETYPE_ROM))
            return FALSE;
    }

    /* loop over regions, then over files */
    for (region = rom_first_region(gamedrv); region; region = rom_next_region(region))
    {
        for (rom = rom_first_file(region); rom; rom = rom_next_file(rom))
        {
            const struct GameDriver *drv;

            name = ROM_GETNAME(rom);
            icrc = ROM_GETCRC(rom);
            length = 0;

            /* obtain CRC-32 and length of ROM file */
            drv = gamedrv;
            do
            {
                err = osd_fchecksum(drv->name, name, &length, &icrc);
                drv = drv->clone_of;
            } while (err && drv && drv != DRIVER_ROOT);

            if (err)
                return FALSE;
        }
    }

    return TRUE;
}


void InitDefaultPropertyPage(HINSTANCE hInst, HWND hWnd)
{
    PROPSHEETHEADER pshead;
    PROPSHEETPAGE   pspage[NUM_PROPSHEETS];
    int             i;
    int             maxPropSheets;

    iGame = -1;
    maxPropSheets = NUM_PROPSHEETS - 2;

    /* Get default options to populate property sheets */
    lpGameOpts = GetDefaultOptions();
    bUseDefaults = FALSE;
    /* Stash the result for comparing later */
    memcpy(&origGameOpts, lpGameOpts, sizeof(options_type));
    bReset = FALSE;
    BuildDataMap();

    ZeroMemory(&pshead, sizeof(PROPSHEETHEADER));
    ZeroMemory(pspage, sizeof(PROPSHEETPAGE) * maxPropSheets);

    /* Fill in the property sheet header */
    pshead.hwndParent     = hWnd;
    pshead.dwSize         = sizeof(PROPSHEETHEADER);
    pshead.dwFlags        = PSH_PROPSHEETPAGE | PSH_USEICONID |
                            PSH_PROPTITLE;
    pshead.hInstance      = hInst;
    pshead.pszCaption     = win32_getstring(WIN32_property_defaultgame);
    pshead.nPages         = maxPropSheets;
    pshead.nStartPage     = 0;
    pshead.pszIcon        = MAKEINTRESOURCE(IDI_MAME32_ICON);
    pshead.ppsp           = pspage;

    /* Fill out the property page templates */
    for (i = 0; i < maxPropSheets; i++)
    {
        pspage[i].dwSize      = sizeof(PROPSHEETPAGE);
        pspage[i].dwFlags     = 0;
        pspage[i].hInstance   = hInst;
        pspage[i].pszTemplate = MAKEINTRESOURCE(dwDlgId[i + 2]);
        pspage[i].pfnCallback = NULL;
        pspage[i].lParam      = 0;
        pspage[i].pfnDlgProc  = GameDisplayOptionsProc;
    }

    /* Create the Property sheet and display it */
    if (PropertySheet(&pshead) == -1)
    {
        char temp[100];
        DWORD dwError = GetLastError();
        sprintf(temp, "%s %d %X", win32_getstring(WIN32_property_sheeterror), dwError, dwError);
        MessageBox(NULL, temp, win32_getstring(WIN32_property_sheeterrortitle), IDOK);
    }
}


void InitPropertyPage(HINSTANCE hInst, HWND hWnd, int game_num)
{
    InitPropertyPageToPage(hInst,hWnd,game_num,PROPERTIES_PAGE);
}


void InitPropertyPageToPage(HINSTANCE hInst, HWND hWnd, int game_num,int start_page)
{
    PROPSHEETHEADER pshead;
    PROPSHEETPAGE   pspage[NUM_PROPSHEETS];
    int             i;
    int             maxPropSheets;

    InitGameAudit(game_num);
    iGame = game_num;
    maxPropSheets = NUM_PROPSHEETS;

    neogeo_set_driver(game_num);

    /* Get Game options to populate property sheets */
    lpGameOpts = GetGameOptions(game_num);
    bUseDefaults = lpGameOpts->use_default;
    /* Stash the result for comparing later */
    memcpy(&origGameOpts, lpGameOpts, sizeof(options_type));
    bReset = FALSE;
    BuildDataMap();

    ZeroMemory(&pshead, sizeof(PROPSHEETHEADER));
    ZeroMemory(pspage, sizeof(PROPSHEETPAGE) * maxPropSheets);

    /* Fill in the property sheet header */
    pshead.hwndParent     = hWnd;
    pshead.dwSize         = sizeof(PROPSHEETHEADER);
    pshead.dwFlags        = PSH_PROPSHEETPAGE | PSH_USEICONID |
                            PSH_PROPTITLE;
    pshead.hInstance      = hInst;
#ifdef BUILD_ENGLISH
    pshead.pszCaption     = ModifyThe(drivers[iGame]->description);
#else
    if (UseJapaneseList())
        pshead.pszCaption     = drivers[iGame]->description;
    else
	    pshead.pszCaption     = ModifyThe(drivers[iGame]->descriptione);
#endif
    pshead.nPages         = maxPropSheets;
    pshead.nStartPage     = start_page;
    pshead.pszIcon        = MAKEINTRESOURCE(IDI_MAME32_ICON);
    pshead.ppsp           = pspage;

    /* Fill out the property page templates */
    for (i = 0; i < maxPropSheets; i++)
    {
        pspage[i].dwSize      = sizeof(PROPSHEETPAGE);
        pspage[i].dwFlags     = 0;
        pspage[i].hInstance   = hInst;
        pspage[i].pszTemplate = MAKEINTRESOURCE(dwDlgId[i]);
        pspage[i].pfnCallback = NULL;
        pspage[i].lParam      = 0;
    }

    pspage[0].pfnDlgProc  = GamePropertiesDialogProc;
    pspage[1].pfnDlgProc  = GameAuditDialogProc;

    pspage[2].pfnDlgProc  = GameDisplayOptionsProc;
    pspage[3].pfnDlgProc  = GameDisplayOptionsProc;
    pspage[4].pfnDlgProc  = GameDisplayOptionsProc;
    pspage[5].pfnDlgProc  = GameDisplayOptionsProc;
    pspage[6].pfnDlgProc  = GameDisplayOptionsProc;

    /* If this is a vector game, add the vector prop sheet */
    if ( maxPropSheets == NUM_PROPSHEETS)
    {
        pspage[NUM_PROPSHEETS - 1].pfnDlgProc  = GameDisplayOptionsProc;
    }

    /* Create the Property sheet and display it */
    if (PropertySheet(&pshead) == -1)
    {
        char temp[100];
        DWORD dwError = GetLastError();
        sprintf(temp, "%s %d %X", win32_getstring(WIN32_property_sheeterror), dwError, dwError);
        MessageBox(NULL, temp, win32_getstring(WIN32_property_sheeterrortitle), IDOK);
    }
}

/*********************************************************************
 * Local Functions
 *********************************************************************/

/* Build CPU info string */
char *GameInfoCPU(UINT nIndex)
{
    int i;
    static char buf[1024] = "";

    ZeroMemory(buf, sizeof(buf));
    i = 0;
    while (i < MAX_CPU && drivers[nIndex]->drv->cpu[i].cpu_type)
    {
        sprintf(&buf[strlen(buf)],"%s %d.%06d MHz",
                cputype_name(drivers[nIndex]->drv->cpu[i].cpu_type),
                drivers[nIndex]->drv->cpu[i].cpu_clock / 1000000,
                drivers[nIndex]->drv->cpu[i].cpu_clock % 1000000);

        if (drivers[nIndex]->drv->cpu[i].cpu_type & CPU_AUDIO_CPU)
            sprintf(&buf[strlen(buf)], win32_getstring(WIN32_property_sound));

        strcat(buf, "\n");

        i++;
    }

    return buf;
}

/* Build Sound system info string */
char *GameInfoSound(UINT nIndex)
{
    int i;
    static char buf[1024] = "";
    const struct MachineDriver *x_driver = drivers[nIndex]->drv;
    const struct MachineSound *x_sound = x_driver->sound;

    ZeroMemory(buf, sizeof(buf));
    for (i = 0; i < MAX_SOUND; i++)
    {
        if (sound_num(&x_sound[i]))
            sprintf(&buf[strlen(buf)],"%d x ", sound_num(&x_sound[i]));

        sprintf(&buf[strlen(buf)],"%s", sound_name(&x_sound[i]));

        if (sound_num(&x_sound[i]))
            sprintf(&buf[strlen(buf)]," %d.%06d MHz",
                sound_clock(&x_sound[i]) / 1000000,
                sound_clock(&x_sound[i]) % 1000000);

        strcat(buf,"\n");
    }
    return buf;
}

/* Build Display info string */
char *GameInfoScreen(UINT nIndex)
{
    static char buf[1024];

    sprintf(buf,"%d x %d (H) %5.2f Hz",
            drivers[nIndex]->drv->default_visible_area.max_x - drivers[nIndex]->drv->default_visible_area.min_x + 1,
            drivers[nIndex]->drv->default_visible_area.max_y - drivers[nIndex]->drv->default_visible_area.min_y + 1,
            drivers[nIndex]->drv->frames_per_second);

    return buf;
}

/* Build color information string */
char *GameInfoColors(UINT nIndex)
{
    static char buf[256];

    sprintf(buf, win32_getstring(WIN32_property_colors), drivers[nIndex]->drv->total_colors);

    return buf;
}

/* Build game status string */
char *GameInfoStatus(UINT nIndex)
{
    switch (GetHasRoms(nIndex))
    {
    case 0:
        return (char *)win32_getstring(WIN32_property_rommissing);

    case 1:
        if (drivers[nIndex]->flags & GAME_NOT_WORKING)
            return (char *)win32_getstring(WIN32_property_notworking);
        if (drivers[nIndex]->flags & GAME_WRONG_COLORS)
            return (char *)win32_getstring(WIN32_property_colorwrong);
        if (drivers[nIndex]->flags & GAME_IMPERFECT_COLORS)
            return (char *)win32_getstring(WIN32_property_imperfectcolor);
        else
            return (char *)win32_getstring(WIN32_property_working);

    default:
    case 2:
        return (char *)win32_getstring(WIN32_property_unknown);
    }
}

/* Build game manufacturer string */
char *GameInfoManufactured(UINT nIndex)
{
    static char buf[1024];

    sprintf(buf, "%s %s", drivers[nIndex]->year, drivers[nIndex]->manufacturer);

    return buf;
}

/* Build Game title string */
char *GameInfoTitle(UINT nIndex)
{
    static char buf[1024];

    if (nIndex == -1)
        strcpy(buf, win32_getstring(WIN32_property_globaloptions));
    else
#ifdef BUILD_ENGLISH
        sprintf(buf,"%s\n\"%s\"", ModifyThe(drivers[nIndex]->description), drivers[nIndex]->name); 
#else
        if (UseJapaneseList())
		{
			sprintf(buf, "%s\n\"%s\"", drivers[nIndex]->description, drivers[nIndex]->name);
		}
		else
		{
			sprintf(buf, "%s\n\"%s\"", ModifyThe(drivers[nIndex]->descriptione), drivers[nIndex]->name);
		}
#endif
    return buf;
}

/* Build game clone infromation string */
char *GameInfoCloneOf(UINT nIndex)
{
    static char buf[1024];

    buf[0] = '\0';

    if (drivers[nIndex]->clone_of != 0
    &&  !(drivers[nIndex]->clone_of->flags & NOT_A_DRIVER))
    {
#ifdef BUILD_ENGLISH
        sprintf(buf, "%s - \"%s\"",
                ModifyThe(drivers[nIndex]->clone_of->description),
                drivers[nIndex]->clone_of->name);
#else
        if (UseJapaneseList())
		{
			sprintf(buf, "%s - \"%s\"",
                drivers[nIndex]->clone_of->description,
                drivers[nIndex]->clone_of->name);
		}
		else
		{
			sprintf(buf, "%s - \"%s\"",
                ModifyThe(drivers[nIndex]->clone_of->descriptione),
                drivers[nIndex]->clone_of->name);
		}
#endif
    }

    return buf;
}

/* Handle the information property page */
static INT_PTR CALLBACK GamePropertiesDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    switch (Msg)
    {
    case WM_INITDIALOG:
#if USE_SINGLELINE_TABCONTROL
        {
            HWND hWnd = PropSheet_GetTabControl(GetParent(hDlg));
            DWORD tabStyle = (GetWindowLong(hWnd,GWL_STYLE) & ~TCS_MULTILINE);
            SetWindowLong(hWnd,GWL_STYLE,tabStyle | TCS_SINGLELINE);
        }
#endif

        Static_SetText(GetDlgItem(hDlg, IDC_PROP_TITLE),         GameInfoTitle(iGame));
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_MANUFACTURED),  GameInfoManufactured(iGame));
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_STATUS),        GameInfoStatus(iGame));
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_CPU),           GameInfoCPU(iGame));
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_SOUND),         GameInfoSound(iGame));
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_SCREEN),        GameInfoScreen(iGame));
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_COLORS),        GameInfoColors(iGame));
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_CLONEOF),       GameInfoCloneOf(iGame));
        if (drivers[iGame]->clone_of != 0
        && !(drivers[iGame]->clone_of->flags & NOT_A_DRIVER))
        {
            ShowWindow(GetDlgItem(hDlg, IDC_PROP_CLONEOF_TEXT), SW_SHOW);
        }
        else
        {
            ShowWindow(GetDlgItem(hDlg, IDC_PROP_CLONEOF_TEXT), SW_HIDE);
        }

        ShowWindow(hDlg,SW_SHOW);
        return 1;
    }
    return 0;
}

BOOL ReadSkipCtrl(HWND hWnd, UINT nCtrlID, int *value)
{
    HWND hCtrl;
    char buf[100];
    int  nValue = *value;

    hCtrl = GetDlgItem(hWnd, nCtrlID);
    if (hCtrl)
    {
        /* Skip lines control */
        Edit_GetText(hCtrl, buf, 100);
        if (sscanf(buf, "%d", value) != 1)
            *value = 0;
    }
    return (nValue == *value) ? FALSE : TRUE;
}

/* Handle all options property pages */
static INT_PTR CALLBACK GameDisplayOptionsProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    switch (Msg)
    {
    case WM_INITDIALOG:
        /* Fill in the Game info at the top of the sheet */
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_TITLE), GameInfoTitle(iGame));
        InitializeOptions(hDlg);
        InitializeMisc(hDlg);

        PopulateControls(hDlg);
        OptionsToProp(hDlg, lpGameOpts);
        SetPropEnabledControls(hDlg);
        if (iGame == -1)
            ShowWindow(GetDlgItem(hDlg, IDC_USE_DEFAULT), SW_HIDE);
        else
            EnableWindow(GetDlgItem(hDlg, IDC_USE_DEFAULT), (bUseDefaults) ? FALSE : TRUE);

        EnableWindow(GetDlgItem(hDlg, IDC_PROP_RESET), bReset);
        ShowWindow(hDlg,SW_SHOW);

        return 1;

    case WM_HSCROLL:
        /* slider changed */
        HANDLE_WM_HSCROLL(hDlg, wParam, lParam, OptOnHScroll);
        bUseDefaults = FALSE;
        bReset = TRUE;
        EnableWindow(GetDlgItem(hDlg, IDC_USE_DEFAULT), TRUE);
        PropSheet_Changed(GetParent(hDlg), hDlg);
        break;

    case WM_COMMAND:
        {
            /* Below, 'changed' is used to signify the 'Apply'
             * button should be enabled.
             */
            WORD wID         = GET_WM_COMMAND_ID(wParam, lParam);
            HWND hWndCtrl    = GET_WM_COMMAND_HWND(wParam, lParam);
            WORD wNotifyCode = GET_WM_COMMAND_CMD(wParam, lParam);
            BOOL changed     = FALSE;

            switch (wID)
            {
            case IDC_FULLSCREEN:
                changed = ReadControl(hDlg, wID);
                break;

            case IDC_SOUNDTYPE:
                if (wNotifyCode == CBN_SELCHANGE)
                    changed = ReadControl(hDlg, wID);
                break;

            case IDC_SIZES:
            case IDC_FRAMESKIP:
            case IDC_RENDER_FULL:
            case IDC_RENDER_WIN:
            case IDC_DEFAULT_INPUT:
            case IDC_SAMPLERATE:
            case IDC_DIJOYID1:
            case IDC_DIJOYID2:
            case IDC_CORETYPE:
            case IDC_DRIVERTYPE:
            case IDC_BIOSTYPE:
                if (wNotifyCode == CBN_SELCHANGE)
                    changed = TRUE;
                break;

            case IDC_TRIPLE_BUFFER :
                if (Button_GetCheck(GetDlgItem(hDlg, IDC_TRIPLE_BUFFER)))
                {
                    Button_SetCheck(GetDlgItem(hDlg, IDC_VSYNC_FULL), FALSE);
                    Button_Enable(GetDlgItem(hDlg,   IDC_VSYNC_FULL), FALSE);
                }
                else
                {
                    Button_Enable(GetDlgItem(hDlg,   IDC_VSYNC_FULL), TRUE);
                }
                changed = TRUE;
                break;

            case IDC_VSYNC_FULL :
                if (Button_GetCheck(GetDlgItem(hDlg, IDC_VSYNC_FULL)))
                {
                    Button_SetCheck(GetDlgItem(hDlg, IDC_TRIPLE_BUFFER), FALSE);
                    Button_Enable(GetDlgItem(hDlg,   IDC_TRIPLE_BUFFER), FALSE);
                }
                else
                {
                    Button_Enable(GetDlgItem(hDlg,   IDC_TRIPLE_BUFFER), TRUE);
                }
                changed = TRUE;
                break;

            case IDC_VSYNC_WIN :
            case IDC_SYSTEMMEMORY :
            case IDC_PATCH_ROMS :
            case IDC_DISABLE_MMX :
            case IDC_LOG :
            case IDC_SLEEP :
			case IDC_CYCLE_CPU :
			case IDC_NGLO_DISABLED :
                changed = ReadControl(hDlg, wID);
                break;

            case IDC_PROP_RESET:
                if (wNotifyCode != BN_CLICKED)
                    break;

                memcpy(lpGameOpts, &origGameOpts, sizeof(options_type));
                BuildDataMap();
                PopulateControls(hDlg);
                OptionsToProp(hDlg, lpGameOpts);
                SetPropEnabledControls(hDlg);
                bReset = FALSE;
                PropSheet_UnChanged(GetParent(hDlg), hDlg);
                bUseDefaults = lpGameOpts->use_default;
                EnableWindow(GetDlgItem(hDlg, IDC_USE_DEFAULT), (bUseDefaults) ? FALSE : TRUE);
                break;

            case IDC_USE_DEFAULT :
                if (iGame != -1)
                {
                    lpGameOpts->use_default = TRUE;
                    lpGameOpts = GetGameOptions(iGame);
                    bUseDefaults = lpGameOpts->use_default;
                    BuildDataMap();
                    PopulateControls(hDlg);
                    OptionsToProp(hDlg, lpGameOpts);
                    SetPropEnabledControls(hDlg);
                    if (origGameOpts.use_default != bUseDefaults)
                    {
                        PropSheet_Changed(GetParent(hDlg), hDlg);
                        bReset = TRUE;
                    }
                    else
                    {
                        PropSheet_UnChanged(GetParent(hDlg), hDlg);
                        bReset = FALSE;
                    }
                    EnableWindow(GetDlgItem(hDlg, IDC_USE_DEFAULT), (bUseDefaults) ? FALSE : TRUE);
                }

                break;

            case IDC_DISPLAYS :
                if (wNotifyCode == LBN_SELCHANGE)
                {
                    DisplayChange(hDlg);
                    changed = TRUE;
                }
                break;

            default:
                if (wNotifyCode == BN_CLICKED)
                    changed = TRUE;
            }

            /* Enable the apply button */
            if (changed == TRUE)
            {
                lpGameOpts->use_default = bUseDefaults = FALSE;
                PropSheet_Changed(GetParent(hDlg), hDlg);
                bReset = TRUE;
                EnableWindow(GetDlgItem(hDlg, IDC_USE_DEFAULT), (bUseDefaults) ? FALSE : TRUE);
            }
            SetPropEnabledControls(hDlg);
        }
        break;

    case WM_NOTIFY:
        switch (((NMHDR *) lParam)->code)
        {
        case PSN_SETACTIVE:
            // Initialize the controls.
            PopulateControls(hDlg);
            OptionsToProp(hDlg, lpGameOpts);
            SetPropEnabledControls(hDlg);
            EnableWindow(GetDlgItem(hDlg, IDC_USE_DEFAULT), (bUseDefaults) ? FALSE : TRUE);
            break;

        case PSN_APPLY:
            // Save and apply the options here
            PropToOptions(hDlg, lpGameOpts);
            ReadControls(hDlg);
            lpGameOpts->use_default = bUseDefaults;
            if (iGame == -1)
                lpGameOpts = GetDefaultOptions();
            else
                lpGameOpts = GetGameOptions(iGame);

            memcpy(&origGameOpts, lpGameOpts, sizeof(options_type));
            BuildDataMap();
            PopulateControls(hDlg);
            OptionsToProp(hDlg, lpGameOpts);
            SetPropEnabledControls(hDlg);
            EnableWindow(GetDlgItem(hDlg, IDC_USE_DEFAULT), (bUseDefaults) ? FALSE : TRUE);
            bReset = FALSE;
            PropSheet_UnChanged(GetParent(hDlg), hDlg);
            SetWindowLong(hDlg, DWL_MSGRESULT, TRUE);
            break;

        case PSN_KILLACTIVE:
            // Save Changes to the options here.
            ReadControls(hDlg);
            ResetDataMap();
            lpGameOpts->use_default = bUseDefaults;
            PropToOptions(hDlg, lpGameOpts);
            SetWindowLong(hDlg, DWL_MSGRESULT, FALSE);
            return 1;

        case PSN_RESET:
            // Reset to the original values.
            // Disregard changes
            memcpy(lpGameOpts, &origGameOpts, sizeof(options_type));
            SetWindowLong(hDlg, DWL_MSGRESULT, FALSE);
            break;

        case PSN_HELP:
            // User wants help for this property page
            break;
        }
        break;
    }
    EnableWindow(GetDlgItem(hDlg, IDC_PROP_RESET), bReset);

    return 0;
}

/* Read controls that are not handled in the DataMap */
void PropToOptions(HWND hWnd, options_type *o)
{
    char buf[100];
    HWND hCtrl;

    o->use_default = bUseDefaults;

    /* Game Display Options box */
    hCtrl = GetDlgItem(hWnd, IDC_SIZES);
    if (hCtrl)
    {
        int index;

        /* Screen size control */
        index = ComboBox_GetCurSel(hCtrl);
        o->display_best_fit = (index > 0) ? FALSE : TRUE;

        if (! o->display_best_fit)
        {
            DWORD   w, h;

            ComboBox_GetText(hCtrl, buf, 100);
            if (sscanf(buf, "%d x %d", &w, &h) == 2)
            {
                o->width  = w;
                o->height = h;
            }
            else
            {
                o->display_best_fit = TRUE;
            }
        }
    }
}

/* Populate controls that are not handled in the DataMap */
void OptionsToProp(HWND hWnd, options_type *o)
{
    HWND hCtrl;
    char buf[100];

    hCtrl = GetDlgItem(hWnd, IDC_SIZES);
    if (hCtrl)
    {
        /* Screen size drop down list */
        if (o->display_best_fit)
        {
            ComboBox_SetCurSel(hCtrl, 0);
        }
        else
        {
            /* Select the mode in the list. */
            int nSelection = 0;
            int nCount = 0;

            /* Get the number of items in the control */
            nCount = ComboBox_GetCount(hCtrl);

            while (0 < nCount--)
            {
                int     nWidth, nHeight;

                /* Get the screen size */
                ComboBox_GetLBText(hCtrl, nCount, buf);

                if (sscanf(buf, "%d x %d", &nWidth, &nHeight) == 2)
                {
                    /* If we match, set nSelection to the right value */
                    if (o->width  == nWidth &&  o->height == nHeight)
                    {
                        nSelection = nCount;
                        break;
                    }
                }
            }
            ComboBox_SetCurSel(hCtrl, nSelection);
        }
    }
    /* Gamma */
    hCtrl = GetDlgItem(hWnd, IDC_GAMMADISP);
    if (hCtrl)
    {
        /* Set the static display to the new value */
        sprintf(buf, "%03.02f", o->gamma);
        Static_SetText(hCtrl, buf);
    }

    hCtrl = GetDlgItem(hWnd, IDC_BRIGHTNESSDISP);
    if (hCtrl)
    {
        /* Set the static display to the new value */
        sprintf(buf, "%3d%%", o->brightness);
        Static_SetText(hCtrl, buf);
    }

    hCtrl = GetDlgItem(hWnd, IDC_VOLUMEDISP);
    if (hCtrl)
    {
        /* Set the static display to the new value */
        sprintf(buf, "%ddB", - o->volume);
        Static_SetText(hCtrl, buf);
    }

    hCtrl = GetDlgItem(hWnd, IDC_DSBUFFER_DISP);
    if (hCtrl)
    {
        /* Set the static display to the new value */
        sprintf(buf, "%dms", o->dsound_buffer);
        Static_SetText(hCtrl, buf);
    }

    hCtrl = GetDlgItem(hWnd, IDC_RANDOMTIMEDISP);
    if (hCtrl)
    {
        /* Set the static display to the new value */
        sprintf(buf, "%ds", o->randomtime);
        Static_SetText(hCtrl, buf);
    }

    hCtrl = GetDlgItem(hWnd, IDC_TRANSDISP);
    if (hCtrl)
    {
        /* Set the static display to the new value */
        sprintf(buf, "%3d%%", o->transparency);
        Static_SetText(hCtrl, buf);
    }

    /* update list of modes for this display by simulating a display change */
    DisplayChange(hWnd);

}

/* Adjust controls - tune them to the currently selected game */
void SetPropEnabledControls(HWND hWnd)
{
    HWND hCtrl;
    int  nIndex;
    int  sound;

    nIndex = iGame;

    ComboBox_Enable(GetDlgItem(hWnd, IDC_DISPLAYS), (DirectDraw_GetNumDisplays() > 1));

    if (nIndex == -1 || Joystick.Available(1) || GameUsesTrackball(nIndex))
        EnableWindow(GetDlgItem(hWnd, IDC_INPUTDEVTEXT), TRUE);
    else
        EnableWindow(GetDlgItem(hWnd, IDC_INPUTDEVTEXT), FALSE);

    hCtrl = GetDlgItem(hWnd, IDC_JOYSTICK);
    if (hCtrl)
        Button_Enable(hCtrl, Joystick.Available(1));

    hCtrl = GetDlgItem(hWnd, IDC_DIJOYSTICK);
    if (hCtrl)
        Button_Enable(hCtrl, DIJoystick.Available(1));

    hCtrl = GetDlgItem(hWnd, IDC_DIKEYBOARD);
    if (hCtrl)
        Button_Enable(hCtrl, DIKeyboard_Available());

    /* Trackball / Mouse options */
    if (nIndex == -1 || GameUsesTrackball(nIndex))
    {
        Button_Enable(GetDlgItem(hWnd, IDC_AI_MOUSE),    TRUE);
        Button_Enable(GetDlgItem(hWnd, IDC_AI_JOYSTICK), TRUE);
        EnableWindow(GetDlgItem(hWnd, IDC_TRACKSPINTEXT),TRUE);
    }
    else
    {
        Button_Enable(GetDlgItem(hWnd, IDC_AI_MOUSE),    FALSE);
        Button_Enable(GetDlgItem(hWnd, IDC_AI_JOYSTICK), FALSE);
        EnableWindow(GetDlgItem(hWnd, IDC_TRACKSPINTEXT),FALSE);
    }

    hCtrl = GetDlgItem(hWnd, IDC_AI_MOUSE);
    if (hCtrl)
    {
        if (Button_GetCheck(hCtrl))
            Button_SetCheck(GetDlgItem(hWnd, IDC_AI_JOYSTICK), FALSE);
        else
            Button_SetCheck(GetDlgItem(hWnd, IDC_AI_JOYSTICK), TRUE);
    }

    /* Sound options */
    hCtrl = GetDlgItem(hWnd, IDC_SOUNDTYPE);
    if (hCtrl)
    {
        sound = ComboBox_GetCurSel(hCtrl);
        ComboBox_Enable(GetDlgItem(hWnd, IDC_SAMPLERATE),  (sound != 0));

        EnableWindow(GetDlgItem(hWnd, IDC_VOLUME),          (sound != 0));
        EnableWindow(GetDlgItem(hWnd, IDC_RATETEXT),        (sound != 0));
        EnableWindow(GetDlgItem(hWnd, IDC_VOLUMEDISP),      (sound != 0));
        EnableWindow(GetDlgItem(hWnd, IDC_DSBUFFER),        (sound == 2));
        EnableWindow(GetDlgItem(hWnd, IDC_DSBUFFER_DISP),   (sound == 2));
        EnableWindow(GetDlgItem(hWnd, IDC_RANDOMTIME),      (sound != 0));
        EnableWindow(GetDlgItem(hWnd, IDC_RANDOMTIMEDISP),  (sound != 0));
    }

    /* Triple Buffer */
    if (Button_GetCheck(GetDlgItem(hWnd, IDC_TRIPLE_BUFFER)))
    {
        Button_SetCheck(GetDlgItem(hWnd, IDC_VSYNC_FULL), FALSE);
        Button_Enable(GetDlgItem(hWnd,   IDC_VSYNC_FULL), FALSE);
    }

    /* Vsync Full Screen */
    if (Button_GetCheck(GetDlgItem(hWnd, IDC_VSYNC_FULL)))
    {
        Button_SetCheck(GetDlgItem(hWnd, IDC_TRIPLE_BUFFER), FALSE);
        Button_Enable(GetDlgItem(hWnd,   IDC_TRIPLE_BUFFER), FALSE);
    }

    /* Turn off MMX if it's not detected */
#ifdef ENABLE_MMX_RENDER
    if (MAME32App.m_bMMXDetected)
    {
        Button_Enable(GetDlgItem(hWnd, IDC_DISABLE_MMX), TRUE);
    }
    else
#endif
    {
        Button_SetCheck(GetDlgItem(hWnd, IDC_DISABLE_MMX), TRUE);
        Button_Enable(GetDlgItem(hWnd, IDC_DISABLE_MMX), FALSE);
    }

    ComboBox_Enable(GetDlgItem(hWnd, IDC_DISPLAYTYPE), TRUE);

    if (nIndex != -1 && GameUsesRomPatch(nIndex))
    {
        Button_Enable(GetDlgItem(hWnd, IDC_PATCH_ROMS), TRUE);
    }
    else
    {
        Button_SetCheck(GetDlgItem(hWnd, IDC_PATCH_ROMS), FALSE);
        Button_Enable(GetDlgItem(hWnd, IDC_PATCH_ROMS), FALSE);
    }

	if (sound != 0)
	{
	    SetStereoEnabled(hWnd, nIndex);
	}
	else
	{
		Button_Enable(GetDlgItem(hWnd, IDC_STEREO), FALSE);
	}

	Button_Enable(GetDlgItem(hWnd, IDC_NGLO_DISABLED), TRUE);
	Button_Enable(GetDlgItem(hWnd, IDC_DIRECTSOUND_FIX), (sound == 2));
}

/**************************************************************
 * Control Helper functions for data exchange
 **************************************************************/
void AssignSampleRate(HWND hWnd)
{
    switch (nSampleRate)
    {
        case 0:  lpGameOpts->sample_rate = 11025; break;
        case 1:  lpGameOpts->sample_rate = 22050; break;
        case 2:  lpGameOpts->sample_rate = 44100; break;
        case 3:  lpGameOpts->sample_rate = 48000; break;
        default: lpGameOpts->sample_rate = 22050; break;
    }
}

void AssignVolume(HWND hWnd)
{
    lpGameOpts->volume = 32 - nVolume;
}

void AssignFullScreen(HWND hWnd)
{
    lpGameOpts->is_window = (bFullScreen) ? FALSE : TRUE;
}

void AssignDSoundBuffer(HWND hWnd)
{
    lpGameOpts->dsound_buffer = nDSBuffer;
}

void AssignRandomTime(HWND hWnd)
{
    lpGameOpts->randomtime = nRandomTime * 5;
}

void AssignGamma(HWND hWnd)
{
    /* Gamma control */
    lpGameOpts->gamma = nGamma / 20.0 + 0.5;
}

/************************************************************
 * DataMap initializers
 ************************************************************/

/* Initialize local helper variables */
void ResetDataMap(void)
{
#if	1
    switch (lpGameOpts->sample_rate)
    {
	    case 11025: nSampleRate = 0; break;
	    case 22050: nSampleRate = 1; break;
	    case 44100: nSampleRate = 2; break;
	    case 48000: nSampleRate = 3; break;
	    default:    nSampleRate = 1; break;
    }
#else
    nSampleRate = (lpGameOpts->sample_rate / 22050) % 3;
#endif
    nVolume     = 32 - lpGameOpts->volume;
    bFullScreen = (lpGameOpts->is_window) ? FALSE : TRUE;
    nGamma      = (int)((lpGameOpts->gamma - 0.5) * 20.0);
    nDSBuffer   = lpGameOpts->dsound_buffer;
	nRandomTime = lpGameOpts->randomtime / 5;
}

/* Build the control mapping by adding all needed information to the DataMap */
void BuildDataMap(void)
{
    InitDataMap();

    ResetDataMap();

    DataMapAdd(IDC_FULLSCREEN,    DM_BOOL, CT_BUTTON,   &bFullScreen,  0, 0, AssignFullScreen);
    DataMapAdd(IDC_SAMPLERATE,    DM_INT,  CT_COMBOBOX, &nSampleRate,  0, 0, AssignSampleRate);
    DataMapAdd(IDC_VOLUME,        DM_INT,  CT_SLIDER,   &nVolume,      0, 0, AssignVolume);
    DataMapAdd(IDC_DSBUFFER,      DM_INT,  CT_SLIDER,   &nDSBuffer,    0, 0, AssignDSoundBuffer);
    DataMapAdd(IDC_RANDOMTIME,    DM_INT,  CT_SLIDER,   &nRandomTime,  0, 0, AssignRandomTime);
    DataMapAdd(IDC_GAMMA,         DM_INT,  CT_SLIDER,   &nGamma,       0, 0, AssignGamma);

    DataMapAdd(IDC_DISPLAYS,        DM_INT,  CT_COMBOBOX, &lpGameOpts->display_monitor,   0, 0, 0);
    DataMapAdd(IDC_TRIPLE_BUFFER,   DM_BOOL, CT_BUTTON,   &lpGameOpts->triple_buffer,     0, 0, 0);
    DataMapAdd(IDC_VSYNC_FULL,      DM_BOOL, CT_BUTTON,   &lpGameOpts->vsync_fullscreen,  0, 0, 0);
    DataMapAdd(IDC_VSYNC_WIN,       DM_BOOL, CT_BUTTON,   &lpGameOpts->vsync_window,      0, 0, 0);
    DataMapAdd(IDC_SYSTEMMEMORY,    DM_BOOL, CT_BUTTON,   &lpGameOpts->systemmemory,      0, 0, 0);
    DataMapAdd(IDC_AI_MOUSE,        DM_BOOL, CT_BUTTON,   &lpGameOpts->use_ai_mouse,      0, 0, 0);
    DataMapAdd(IDC_DIKEYBOARD,      DM_BOOL, CT_BUTTON,   &lpGameOpts->di_keyboard,       0, 0, 0);
    DataMapAdd(IDC_DIJOYSTICK,      DM_BOOL, CT_BUTTON,   &lpGameOpts->di_joystick,       0, 0, 0);
    DataMapAdd(IDC_JOYSTICK,        DM_BOOL, CT_BUTTON,   &lpGameOpts->use_joystick,      0, 0, 0);
    DataMapAdd(IDC_DEFAULT_INPUT,   DM_INT,  CT_COMBOBOX, &lpGameOpts->default_input,     0, 0, 0);
    DataMapAdd(IDC_DIJOYID1,        DM_INT,  CT_COMBOBOX, &lpGameOpts->dijoyid[0],        0, 0, 0);
    DataMapAdd(IDC_DIJOYID2,        DM_INT,  CT_COMBOBOX, &lpGameOpts->dijoyid[1],        0, 0, 0);
    DataMapAdd(IDC_STEREO,          DM_BOOL, CT_BUTTON,   &lpGameOpts->stereo,            0, 0, 0);
    DataMapAdd(IDC_DIRECTSOUND_FIX, DM_BOOL, CT_BUTTON,   &lpGameOpts->directsound_fix,   0, 0, 0);
    DataMapAdd(IDC_CHEAT,           DM_BOOL, CT_BUTTON,   &lpGameOpts->cheat,             0, 0, 0);
    DataMapAdd(IDC_AUTO_PAUSE,      DM_BOOL, CT_BUTTON,   &lpGameOpts->auto_pause,        0, 0, 0);
    DataMapAdd(IDC_DISABLE_MMX,     DM_BOOL, CT_BUTTON,   &lpGameOpts->disable_mmx,       0, 0, 0);
    DataMapAdd(IDC_LOG,             DM_BOOL, CT_BUTTON,   &lpGameOpts->error_log,         0, 0, 0);
    DataMapAdd(IDC_SLEEP,           DM_BOOL, CT_BUTTON,   &lpGameOpts->dontsleep,         0, 0, 0);
    DataMapAdd(IDC_CYCLE_CPU,       DM_BOOL, CT_BUTTON,   &lpGameOpts->cyclecpu,          0, 0, 0);
    DataMapAdd(IDC_RENDER_FULL,     DM_INT,  CT_COMBOBOX, &lpGameOpts->render_fullscreen, 0, 0, 0);
    DataMapAdd(IDC_RENDER_WIN,      DM_INT,  CT_COMBOBOX, &lpGameOpts->render_window,     0, 0, 0);
    DataMapAdd(IDC_CORETYPE,        DM_INT,  CT_COMBOBOX, &lpGameOpts->core_type,         0, 0, 0);
    DataMapAdd(IDC_DRIVERTYPE,      DM_INT,  CT_COMBOBOX, &lpGameOpts->driver_type,       0, 0, 0);
    DataMapAdd(IDC_BIOSTYPE,        DM_INT,  CT_COMBOBOX, &lpGameOpts->bios_type,         0, 0, 0);
    DataMapAdd(IDC_PATCH_ROMS,      DM_BOOL, CT_BUTTON,   &lpGameOpts->patch_roms,        0, 0, 0);
    DataMapAdd(IDC_NGLO_DISABLED,   DM_BOOL, CT_BUTTON,   &lpGameOpts->nglo_disabled,     0, 0, 0);
    DataMapAdd(IDC_SOUNDTYPE,       DM_INT,  CT_COMBOBOX, &lpGameOpts->sound,             0, 0, 0);
    DataMapAdd(IDC_FRAMESKIP,       DM_INT,  CT_COMBOBOX, &lpGameOpts->frame_skip,        0, 0, 0);
    DataMapAdd(IDC_BRIGHTNESS,      DM_INT,  CT_SLIDER,   &lpGameOpts->brightness,        0, 0, 0);
	DataMapAdd(IDC_TRANSPARENT,     DM_INT,  CT_SLIDER,   &lpGameOpts->transparency,      0, 0, 0);
}

static void SetStereoEnabled(HWND hWnd, int index)
{
    BOOL enabled = FALSE;
    HWND hCtrl;

    hCtrl = GetDlgItem(hWnd, IDC_STEREO);
    if (hCtrl)
    {
        if (index == -1 || drivers[index]->drv->sound_attributes & SOUND_SUPPORTS_STEREO)
            enabled = TRUE;

        EnableWindow(hCtrl, enabled);
        if (!enabled)
            Button_SetCheck(hCtrl, FALSE);
    }
}

/* Moved here cause it's called in a few places */
static void InitializeOptions(HWND hDlg)
{
    InitializeDisplayTypeUI(hDlg);
    InitializeDisplayModeUI(hDlg);
    InitializeRender(hDlg);
    InitializeSoundUI(hDlg);
    InitializeSkippingUI(hDlg);
    InitializeDefaultInputUI(hDlg);
    InitializeDIjoyUI(hDlg);
    InitializeCoreUI(hDlg);
    InitializeDriverUI(hDlg);
    InitializeBiosUI(hDlg);
}

/* Moved here because it is called in several places */
static void InitializeMisc(HWND hDlg)
{
    Button_Enable(GetDlgItem(hDlg, IDC_JOYSTICK),   Joystick.Available(1));
    Button_Enable(GetDlgItem(hDlg, IDC_DIJOYSTICK), DIJoystick.Available(1));
    Button_Enable(GetDlgItem(hDlg, IDC_DIKEYBOARD), DIKeyboard_Available());

    SendMessage(GetDlgItem(hDlg, IDC_GAMMA), TBM_SETRANGE,
                (WPARAM)FALSE,
                (LPARAM)MAKELONG(0, 30)); /* 0.50 - 2.00 in .05 increments */

    SendMessage(GetDlgItem(hDlg, IDC_BRIGHTNESS), TBM_SETRANGE,
                (WPARAM)FALSE,
                (LPARAM)MAKELONG(0, 100));

    SendMessage(GetDlgItem(hDlg, IDC_VOLUME), TBM_SETRANGE,
                (WPARAM)FALSE,
                (LPARAM)MAKELONG(0, 32)); /* -32 - 0 */

    SendMessage(GetDlgItem(hDlg, IDC_TRANSPARENT), TBM_SETRANGE,
                (WPARAM)FALSE,
                (LPARAM)MAKELONG(0, 100));

    SendMessage(GetDlgItem(hDlg, IDC_DSBUFFER), TBM_SETRANGE,
                (WPARAM)FALSE,
                (LPARAM)MAKELONG(50, 200)); /* 0 - 200 */

    SendMessage(GetDlgItem(hDlg, IDC_RANDOMTIME), TBM_SETRANGE,
                (WPARAM)FALSE,
                (LPARAM)MAKELONG(1, 60));
}

static void OptOnHScroll(HWND hwnd, HWND hwndCtl, UINT code, int pos)
{
    if (hwndCtl == GetDlgItem(hwnd, IDC_GAMMA))
    {
        GammaSelectionChange(hwnd);
    }
    else
    if (hwndCtl == GetDlgItem(hwnd, IDC_BRIGHTNESS))
    {
        BrightnessSelectionChange(hwnd);
    }
    else
    if (hwndCtl == GetDlgItem(hwnd, IDC_VOLUME))
    {
        VolumeSelectionChange(hwnd);
    }
    else
    if (hwndCtl == GetDlgItem(hwnd, IDC_DSBUFFER))
    {
        DSoundBufferSelectionChange(hwnd);
    }
    else
    if (hwndCtl == GetDlgItem(hwnd, IDC_RANDOMTIME))
    {
        RandomTimeChange(hwnd);
    }
    else
	if (hwndCtl == GetDlgItem(hwnd, IDC_TRANSPARENT))
	{
		TransparentChange(hwnd);
	}
}


/* Handle changes to the Gamma slider */
static void GammaSelectionChange(HWND hwnd)
{
    char buf[100];
    UINT nValue;
    double dGamma;

    /* Get the current value of the control */
    nValue = SendMessage(GetDlgItem(hwnd, IDC_GAMMA), TBM_GETPOS, 0, 0);

    dGamma = nValue / 20.0 + 0.5;

    /* Set the static display to the new value */
    sprintf(buf, "%03.02f", dGamma);
    Static_SetText(GetDlgItem(hwnd, IDC_GAMMADISP), buf);
}

/* Handle changes to the Brightness slider */
static void BrightnessSelectionChange(HWND hwnd)
{
    char buf[100];
    int  nValue;

    /* Get the current value of the control */
    nValue = SendMessage(GetDlgItem(hwnd, IDC_BRIGHTNESS), TBM_GETPOS, 0, 0);

    /* Set the static display to the new value */
    sprintf(buf, "%3d%%", nValue);
    Static_SetText(GetDlgItem(hwnd, IDC_BRIGHTNESSDISP), buf);
}

/* Handle changes to the Brightness slider */
static void TransparentChange(HWND hwnd)
{
    char buf[100];
    int  nValue;

    /* Get the current value of the control */
    nValue = SendMessage(GetDlgItem(hwnd, IDC_TRANSPARENT), TBM_GETPOS, 0, 0);

    /* Set the static display to the new value */
    sprintf(buf, "%3d%%", nValue);
    Static_SetText(GetDlgItem(hwnd, IDC_TRANSDISP), buf);
}

/* Handle changes to the current display to use */
static void DisplayChange(HWND hWnd)
{
    int     nCurSelection;

    nCurSelection = ComboBox_GetCurSel(GetDlgItem(hWnd, IDC_DISPLAYS));
    if (nCurSelection != CB_ERR)
    {
       /* reset modes with the newly selected display */
       DirectDraw_CreateByIndex(nCurSelection);
       UpdateDisplayModeUI(hWnd, 16);
    }
}

/* Handle changes to the Volume slider */
static void VolumeSelectionChange(HWND hwnd)
{
    char buf[100];
    int  nValue;

    /* Get the current value of the control */
    nValue = SendMessage(GetDlgItem(hwnd, IDC_VOLUME), TBM_GETPOS, 0, 0);

    /* Set the static display to the new value */
    sprintf(buf, "%ddB", nValue - 32);
    Static_SetText(GetDlgItem(hwnd, IDC_VOLUMEDISP), buf);
}

/* Handle changes to the DirectSoundBuffer slider */
static void DSoundBufferSelectionChange(HWND hwnd)
{
    char buf[100];
    int  nValue;

    /* Get the current value of the control */
    nValue = SendMessage(GetDlgItem(hwnd, IDC_DSBUFFER), TBM_GETPOS, 0, 0);
    
    /* Set the static display to the new value */
    sprintf(buf, "%3dms", nValue);
    Static_SetText(GetDlgItem(hwnd, IDC_DSBUFFER_DISP), buf);
}

/* Handle changes to the Random Time slider */
static void RandomTimeChange(HWND hwnd)
{
    char buf[100];
    int  nValue;

    /* Get the current value of the control */
    nValue = SendMessage(GetDlgItem(hwnd, IDC_RANDOMTIME), TBM_GETPOS, 0, 0);

    /* Set the static display to the new value */
    sprintf(buf, "%ds", nValue * 5);
    Static_SetText(GetDlgItem(hwnd, IDC_RANDOMTIMEDISP), buf);
}

/* Adjust possible choices in the Screen Size drop down */
static void UpdateDisplayModeUI(HWND hwnd, DWORD dwDepth)
{
    int                     i;
    char                    buf[100];
    struct tDisplayModes*   pDisplayModes;
    int                     nPick;
    int                     nCount = 0;
    int                     nSelection = 0;
    DWORD                   w = 0, h = 0;
    HWND                    hCtrl = GetDlgItem(hwnd, IDC_SIZES);

    if (!hCtrl)
        return;

    /* Find out what is currently selected if anything. */
    nPick = ComboBox_GetCurSel(hCtrl);
    if (nPick != 0 && nPick != CB_ERR)
    {
        ComboBox_GetText(GetDlgItem(hwnd, IDC_SIZES), buf, 100);
        if (sscanf(buf, "%d x %d", &w, &h) != 2)
        {
            w = 0;
            h = 0;
        }
    }

    dwDepth = 16;

    /* Remove all items in the list. */
    ComboBox_ResetContent(hCtrl);

    ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_autosize));

    pDisplayModes = DirectDraw_GetDisplayModes();

    for (i = 0; i < pDisplayModes->m_nNumModes; i++)
    {
        if (pDisplayModes->m_Modes[i].m_dwBPP == dwDepth)
        {
            nCount++;
            sprintf(buf, "%i x %i", pDisplayModes->m_Modes[i].m_dwWidth,
                                    pDisplayModes->m_Modes[i].m_dwHeight);
            ComboBox_AddString(hCtrl, buf);

            if (w == pDisplayModes->m_Modes[i].m_dwWidth
            &&  h == pDisplayModes->m_Modes[i].m_dwHeight)
                nSelection = nCount;
        }
    }

    ComboBox_SetCurSel(hCtrl, nSelection);
}

/* Initialize the Display options to auto mode */
static void InitializeDisplayModeUI(HWND hwnd)
{
    UpdateDisplayModeUI(hwnd, 0);
}

/* Initialize the sound options */
static void InitializeSoundUI(HWND hwnd)
{
    int     sound;
    HWND    hCtrl;

    sound = SOUND_NONE;

    hCtrl = GetDlgItem(hwnd, IDC_SOUNDTYPE);
    if (hCtrl)
    {
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_nosound));
#ifndef NOMIDAS
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_midas));
#endif
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_directsound));

        ComboBox_SetCurSel(GetDlgItem(hwnd, IDC_SOUNDTYPE), sound);
    }

    hCtrl = GetDlgItem(hwnd, IDC_SAMPLERATE);
    if (hCtrl)
    {
        ComboBox_AddString(hCtrl, "11025");
        ComboBox_AddString(hCtrl, "22050");
        ComboBox_AddString(hCtrl, "44100");
        ComboBox_AddString(hCtrl, "48000");
        ComboBox_SetCurSel(hCtrl, 1);
    }
}

#ifdef MULTI_MONITOR

/* Populate the Display type drop down */
static void InitializeDisplayTypeUI(HWND hwnd)
{
    int i;

    HWND hWndDisplays = GetDlgItem(hwnd, IDC_DISPLAYS);

    if (hWndDisplays == 0)
        return;

    ComboBox_ResetContent(hWndDisplays);
    for (i = 0; i < DirectDraw_GetNumDisplays(); i++)
    {
        ComboBox_AddString(hWndDisplays, DirectDraw_GetDisplayName(i));
    }
}

#else

/* Populate the Display type drop down */
static void InitializeDisplayTypeUI(HWND hwnd)
{
    hWndDisplays = GetDlgItem(hwnd, IDC_DISPLAYS);

    if (hWndDisplays == 0)
        return;

    ComboBox_ResetContent(hWndDisplays);
    ComboBox_AddString(hWndDisplays, win32_getstring(WIN32_property_primarydisplay));
    ComboBox_SetCurSel(hWndDisplays, 0);
    nNumDisplays++;
}

#endif

/* Populate the Frame Skipping drop down */
static void InitializeRender(HWND hwnd)
{
    HWND hCtrl = GetDlgItem(hwnd, IDC_RENDER_FULL);

    if (hCtrl)
    {
		int i;

		for (i = 0; i < 9; i++)
			ComboBox_AddString(hCtrl, win32_getstring(i + WIN32_property_x1));
    }

    hCtrl = GetDlgItem(hwnd, IDC_RENDER_WIN);

    if (hCtrl)
    {
		int i;

		for (i = 0; i < 8; i++)
			ComboBox_AddString(hCtrl, win32_getstring(i + WIN32_property_x1));
    }
}

/* Populate the Frame Skipping drop down */
static void InitializeSkippingUI(HWND hwnd)
{
    HWND hCtrl = GetDlgItem(hwnd, IDC_FRAMESKIP);

    if (hCtrl)
    {
        int i;
        char buf[64];

        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_autoframeskip));
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_everyframe));

        for (i = 1; i < 12; i++)
        {
            sprintf(buf, win32_getstring(WIN32_property_frameskip), i);
            ComboBox_AddString(hCtrl, buf);
        }
    }
}

/* Populate the Default Input drop down */
static void InitializeDefaultInputUI(HWND hwnd)
{
    HWND hCtrl = GetDlgItem(hwnd, IDC_DEFAULT_INPUT);

    if (hCtrl)
    {
		int i;

		for (i = 0; i < 3; i++)
			ComboBox_AddString(hCtrl, win32_getstring(i + WIN32_property_normal));
    }
}

static void InitializeDIjoyUI(HWND hwnd)
{
    int i;
    char buf[8];
    HWND hCtrl;

    hCtrl = GetDlgItem(hwnd, IDC_DIJOYID1);
    if (hCtrl)
    {
        for (i = 0; i < 16; i++)
        {
            sprintf(buf, "ID%d", i + 1);
            ComboBox_AddString(hCtrl, buf);
        }
    }

    hCtrl = GetDlgItem(hwnd, IDC_DIJOYID2);
    if (hCtrl)
    {
        for (i = 0; i < 16; i++)
        {
            sprintf(buf, "ID%d", i + 1);
            ComboBox_AddString(hCtrl, buf);
        }
    }
}

static void InitializeCoreUI(HWND hwnd)
{
    HWND hCtrl = GetDlgItem(hwnd, IDC_CORETYPE);

    if (hCtrl)
    {
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_ASMcore)); /* 0 */
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_Ccore));   /* 1 */
    }
}

static void InitializeDriverUI(HWND hwnd)
{
    HWND hCtrl = GetDlgItem(hwnd, IDC_DRIVERTYPE);

    if (hCtrl)
    {
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_standarddriver)); /* 0 */
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_rasterdriver));   /* 1 */
    }
}

static void InitializeBiosUI(HWND hwnd)
{
    HWND hCtrl = GetDlgItem(hwnd, IDC_BIOSTYPE);

    if (hCtrl)
    {
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_standardBIOS));   /* 0 */
        ComboBox_AddString(hCtrl, win32_getstring(WIN32_property_debugBIOS));      /* 1 */
    }
}

/**************************************************************************
    Game History / Command List functions
 **************************************************************************/

/* Load indexes from history.dat if found */
char * GameHistory(int game_index)
{
    historyBuf[0] = '\0';

    if (load_driver_history(drivers[game_index], historyBuf, sizeof(historyBuf)) == 0)
        FixBuf(historyBuf);

    return historyBuf;
}

/* Load indexes from command.dat if found */
char * GameCommand(int game_index)
{
    commandBuf[0] = '\0';

    if (load_driver_command(drivers[game_index], commandBuf, sizeof(commandBuf)) == 0)
        FixBuf(commandBuf);

    return commandBuf;
}

static void FixBuf(char *buf)
{
    char *s  = tempConvertBuf;
    char *p  = buf;
    int  len = 0;

    if (strlen(buf) < 3)
    {
        *buf = '\0';
        return;
    }

    while (*p && len < MAX_DATAFILE_LEN - 1)
    {
        if (*p == '\n')
        {
            *s++ = '\r';
            len++;
        }

        *s++ = *p++;
        len++;
    }

    *s++ = '\0';
    strcpy(buf, tempConvertBuf);
}


/* End of source file */
