/*********************************************************************

    common.c

    Generic functions, mostly ROM and graphics related.

*********************************************************************/

#include "driver.h"
#include "png.h"
#include <stdarg.h>
#include "drivers/neogeo.h"


//#define LOG_LOAD


/***************************************************************************

    Type definitions

***************************************************************************/

struct rom_load_data
{
    int    warnings;        /* warning count during processing */
    int    errors;          /* error count during processing   */

    int    romsloaded;      /* current ROMs loaded count       */
    int    romstotal;       /* total number of ROMs to read    */

    void   *file;           /* current file                    */

    UINT8  *regionbase;     /* base of current region          */
    UINT32 regionlength;    /* length of current region        */

    char   errorbuf[4096];  /* accumulated errors              */
    UINT8  tempbuf[65536];  /* temporary buffer                */
};

static disable_crc_check = 0; /* for rom patch */

/***************************************************************************

    Global variables

***************************************************************************/

/* These globals are only kept on a machine basis - LBO 042898 */
unsigned int dispensed_tickets;
unsigned int coins[COIN_COUNTERS];
unsigned int lastcoin[COIN_COUNTERS];
unsigned int coinlockedout[COIN_COUNTERS];


/***************************************************************************

    Prototypes

***************************************************************************/

static int rom_load_new(const struct RomModule *romp);


/***************************************************************************

  Read ROMs into memory.

  Arguments:
  const struct RomModule *romp - pointer to an array of Rommodule structures,
                                 as defined in common.h.

***************************************************************************/

int readroms(void)
{
    return rom_load_new(Machine->gamedrv->rom);
}


/***************************************************************************

  ROM parsing helpers

***************************************************************************/

const struct RomModule *rom_first_region(const struct GameDriver *drv)
{
    return drv->rom;
}

const struct RomModule *rom_next_region(const struct RomModule *romp)
{
    romp++;
    while (!ROMENTRY_ISREGIONEND(romp))
        romp++;
    return ROMENTRY_ISEND(romp) ? NULL : romp;
}

const struct RomModule *rom_first_file(const struct RomModule *romp)
{
    romp++;
    while (!ROMENTRY_ISFILE(romp) && !ROMENTRY_ISREGIONEND(romp))
        romp++;
    return ROMENTRY_ISREGIONEND(romp) ? NULL : romp;
}

const struct RomModule *rom_next_file(const struct RomModule *romp)
{
    romp++;
    while (!ROMENTRY_ISFILE(romp) && !ROMENTRY_ISREGIONEND(romp))
        romp++;
    return ROMENTRY_ISREGIONEND(romp) ? NULL : romp;
}

const struct RomModule *rom_first_chunk(const struct RomModule *romp)
{
    return (ROMENTRY_ISFILE(romp)) ? romp : NULL;
}

const struct RomModule *rom_next_chunk(const struct RomModule *romp)
{
    romp++;
    return (ROMENTRY_ISCONTINUE(romp)) ? romp : NULL;
}


/***************************************************************************

  Memory regions.

***************************************************************************/

unsigned char *memory_region(int num)
{
    int i;

    if (num < MAX_MEMORY_REGIONS)
        return Machine->memory_region[num].base;
    else
    {
        for (i = 0;i < MAX_MEMORY_REGIONS;i++)
        {
            if (Machine->memory_region[i].type == num)
                return Machine->memory_region[i].base;
        }
    }

    return 0;
}

size_t memory_region_length(int num)
{
    int i;

    if (num < MAX_MEMORY_REGIONS)
        return Machine->memory_region[num].length;
    else
    {
        for (i = 0;i < MAX_MEMORY_REGIONS;i++)
        {
            if (Machine->memory_region[i].type == num)
                return Machine->memory_region[i].length;
        }
    }

    return 0;
}

int new_memory_region(int num, size_t length, UINT32 flags)
{
    int i;

    if (num < MAX_MEMORY_REGIONS)
    {
        Machine->memory_region[num].length = length;
        Machine->memory_region[num].base = MyMalloc(length);
        return (Machine->memory_region[num].base == NULL) ? 1 : 0;
    }
    else
    {
        for (i = 0;i < MAX_MEMORY_REGIONS;i++)
        {
            if (Machine->memory_region[i].base == NULL)
            {
                Machine->memory_region[i].length = length;
                Machine->memory_region[i].type = num;
                Machine->memory_region[i].flags = flags;
                Machine->memory_region[i].base = MyMalloc(length);
                return (Machine->memory_region[i].base == NULL) ? 1 : 0;
            }
        }
    }
    return 1;
}

void free_memory_region(int num)
{
    int i;

    if (num < MAX_MEMORY_REGIONS)
    {
        MyFree(Machine->memory_region[num].base);
        memset(&Machine->memory_region[num], 0, sizeof(Machine->memory_region[num]));
    }
    else
    {
        for (i = 0;i < MAX_MEMORY_REGIONS;i++)
        {
            if (Machine->memory_region[i].type == num)
            {
                MyFree(Machine->memory_region[i].base);
                memset(&Machine->memory_region[i], 0, sizeof(Machine->memory_region[i]));
                return;
            }
        }
    }
}

/***************************************************************************


***************************************************************************/

void set_visible_area(int min_x, int max_x, int min_y, int max_y)
{
    Machine->visible_area.min_x = min_x;
    Machine->visible_area.max_x = max_x;
    Machine->visible_area.min_y = min_y;
    Machine->visible_area.max_y = max_y;

    osd_set_visible_area(min_x, max_x, min_y, max_y);
}


struct osd_bitmap *bitmap_alloc(int width, int height)
{
    return osd_alloc_bitmap(width, height);
}


void bitmap_free(struct osd_bitmap *bitmap)
{
    osd_free_bitmap(bitmap);
}


void save_screen_snapshot_as(void *fp, struct osd_bitmap *bitmap)
{
    struct osd_bitmap *copy;
    int sizex, sizey;

    sizex = Machine->visible_area.max_x - Machine->visible_area.min_x + 1;
    sizey = Machine->visible_area.max_y - Machine->visible_area.min_y + 1;

    copy = bitmap_alloc(sizex, sizey);

    if (copy)
    {
        copybitmap(copy,
                bitmap,
                0,
                0,
                -Machine->visible_area.min_x,
                -Machine->visible_area.min_y,
                0,
                TRANSPARENCY_NONE,
				0);
        png_write_bitmap(fp, copy);
        bitmap_free(copy);
    }
}


int snapno = 0;

void save_screen_snapshot(struct osd_bitmap *bitmap)
{
    void *fp;
    char name[20];

    /* avoid overwriting existing files */
    /* first of all try with "gamename.png" */
    sprintf(name,"%s", Machine->gamedrv->name);
    if (osd_faccess(name, OSD_FILETYPE_SCREENSHOT))
    {
        do
        {
            /* otherwise use "name_NN.png" */
            sprintf(name,"%s_%02d", Machine->gamedrv->name, snapno++);
        } while (osd_faccess(name, OSD_FILETYPE_SCREENSHOT));
    }

    if ((fp = osd_fopen(name, 0, OSD_FILETYPE_SCREENSHOT, 1)) != NULL)
    {
        save_screen_snapshot_as(fp, bitmap);
        osd_fclose(fp);
    }
}


/*-------------------------------------------------
    debugload - log data to a file
-------------------------------------------------*/

void CLIB_DECL debugload(const char *string, ...)
{
#ifdef LOG_LOAD
    static int opened;
    va_list arg;
    FILE *f;

    f = fopen("romload.log", opened++ ? "a" : "w");
    if (f)
    {
        va_start(arg, string);
        vfprintf(f, string, arg);
        va_end(arg);
        fclose(f);
    }
#endif
}


/*-------------------------------------------------
    count_roms - counts the total number of ROMs
    that will need to be loaded
-------------------------------------------------*/

static int count_roms(const struct RomModule *romp)
{
    const struct RomModule *region, *rom;
    int count = 0;

    /* loop over regions, then over files */
    for (region = romp; region; region = rom_next_region(region))
        for (rom = rom_first_file(region); rom; rom = rom_next_file(rom))
            count++;

    /* return the total count */
    return count;
}


/*-------------------------------------------------
    fill_random - fills an area of memory with
    random data
-------------------------------------------------*/

static void fill_random(UINT8 *base, UINT32 length)
{
    while (length--)
        *base++ = rand();
}


/*-------------------------------------------------
    handle_missing_file - handles error generation
    for missing files
-------------------------------------------------*/

static void handle_missing_file(struct rom_load_data *romdata, const struct RomModule *romp)
{
    /* no good dumps are okay */
    if (ROM_NOGOODDUMP(romp))
    {
        sprintf(&romdata->errorbuf[strlen(romdata->errorbuf)], "%-12s NOT FOUND (NO GOOD DUMP KNOWN)\n", ROM_GETNAME(romp));
        romdata->warnings++;
    }

    /* anything else is bad */
    else
    {
        sprintf(&romdata->errorbuf[strlen(romdata->errorbuf)], "%-12s NOT FOUND\n", ROM_GETNAME(romp));
        romdata->errors++;
    }
}


/*-------------------------------------------------
    verify_length_and_crc - verify the length
    and CRC of a file
-------------------------------------------------*/

static void verify_length_and_crc(struct rom_load_data *romdata, const char *name, UINT32 explength, UINT32 expcrc)
{
    UINT32 actlength, actcrc;

    /* we've already complained if there is no file */
    if (!romdata->file)
        return;

    /* get the length and CRC from the file */
    actlength = osd_fsize(romdata->file);
    actcrc = osd_fcrc(romdata->file);

    /* verify length */
    if (explength != actlength)
    {
        sprintf(&romdata->errorbuf[strlen(romdata->errorbuf)], "%-12s WRONG LENGTH (expected: %08x found: %08x)\n", name, explength, actlength);
        romdata->warnings++;
    }

    /* verify CRC */
    if (expcrc != actcrc && !disable_crc_check)
    {
        /* expected CRC == 0 means no good dump known */
        if (expcrc == 0)
            sprintf(&romdata->errorbuf[strlen(romdata->errorbuf)], "%-12s NO GOOD DUMP KNOWN\n", name);

        /* inverted CRC means needs redump */
        else if (expcrc == BADCRC(actcrc))
            sprintf(&romdata->errorbuf[strlen(romdata->errorbuf)], "%-12s ROM NEEDS REDUMP\n",name);

        /* otherwise, it's just bad */
        else
            sprintf(&romdata->errorbuf[strlen(romdata->errorbuf)], "%-12s WRONG CRC (expected: %08x found: %08x)\n", name, expcrc, actcrc);
        romdata->warnings++;
    }
}


/*-------------------------------------------------
    display_rom_load_results - display the final
    results of ROM loading
-------------------------------------------------*/

static int display_rom_load_results(struct rom_load_data *romdata)
{
    int region, lastres;
    extern int bailing;

    /* final status display */
    lastres = osd_display_loading_rom_message(NULL, romdata->romsloaded, romdata->romstotal);
    romdata->errors |= lastres;

    /* only display if we have warnings or errors */
    if (romdata->warnings || romdata->errors)
    {
        /* display either an error message or a warning message */
        if (romdata->errors)
        {
            strcat(romdata->errorbuf, "ERROR: required files are missing, the game cannot be run.\n");
            bailing = 1;
        }
        else
            strcat(romdata->errorbuf, "WARNING: the game might not run correctly.\n");

        /* display the result */
        printf("%s", romdata->errorbuf);

        /* if we're not getting out of here, wait for a keypress */
        if (!options.gui_host && !bailing)
        {
            int k;

            /* loop until we get one */
            printf("Press any key to continue\n");
            do
            {
                k = code_read_async();
            }
            while (k == CODE_NONE || k == KEYCODE_LCONTROL);

            /* bail on a control + C */
            if (keyboard_pressed(KEYCODE_LCONTROL) && keyboard_pressed(KEYCODE_C))
                return 1;
        }
    }

    /* clean up any regions */
    if (romdata->errors)
        for (region = 0; region < MAX_MEMORY_REGIONS; region++)
            free_memory_region(region);

    /* return true if we had any errors */
    return (romdata->errors != 0);
}


/*-------------------------------------------------
    region_post_process - post-process a region,
    byte swapping and inverting data as necessary
-------------------------------------------------*/

static void region_post_process(struct rom_load_data *romdata, const struct RomModule *regiondata)
{
    int type = ROMREGION_GETTYPE(regiondata);
    int datawidth = ROMREGION_GETWIDTH(regiondata) / 8;
    int littleendian = ROMREGION_ISLITTLEENDIAN(regiondata);
    UINT8 *base;
    int i, j;

    debugload("+ datawidth=%d little=%d\n", datawidth, littleendian);

    /* if this is a CPU region, override with the CPU width and endianness */
    if (type >= REGION_CPU1 && type < REGION_CPU1 + MAX_CPU)
    {
        int cputype = Machine->drv->cpu[type - REGION_CPU1].cpu_type & ~CPU_FLAGS_MASK;
        if (cputype != 0)
        {
            datawidth = cpuintf[cputype].databus_width / 8;
            littleendian = (cpuintf[cputype].endianess == CPU_IS_LE);
            debugload("+ CPU region #%d: datawidth=%d little=%d\n", type - REGION_CPU1, datawidth, littleendian);
        }
    }

    /* if the region is inverted, do that now */
    if (ROMREGION_ISINVERTED(regiondata))
    {
        debugload("+ Inverting region\n");
        for (i = 0, base = romdata->regionbase; i < romdata->regionlength; i++)
            *base++ ^= 0xff;
    }

    /* swap the endianness if we need to */
#ifdef LSB_FIRST
    if (datawidth > 1 && !littleendian)
#else
    if (datawidth > 1 && littleendian)
#endif
    {
        debugload("+ Byte swapping region\n");
        for (i = 0, base = romdata->regionbase; i < romdata->regionlength; i += datawidth)
        {
            UINT8 temp[8];
            memcpy(temp, base, datawidth);
            for (j = datawidth - 1; j >= 0; j--)
                *base++ = temp[j];
        }
    }
}


/*-------------------------------------------------
    open_rom_file - open a ROM file, searching
    up the parent and loading via CRC
-------------------------------------------------*/
#include <windows.h>

static int open_rom_file(struct rom_load_data *romdata, const struct RomModule *romp)
{
    const struct GameDriver *drv;
    char crc[9];

    /* update status display */
    if (osd_display_loading_rom_message(ROM_GETNAME(romp), ++romdata->romsloaded, romdata->romstotal) != 0)
       return 0;

    /* first attempt reading up the chain through the parents */
    romdata->file = NULL;
    for (drv = Machine->gamedrv; !romdata->file && drv; drv = drv->clone_of)
    {
        const char *name = ROM_GETNAME(romp);

        if (!stricmp(name, "neo-geo.rom"))
        {
            const char *bios_name = NULL;

            if (strcmp(drv->name, "irrmaze"))
            {
                if (options.bios_type == BIOS_TYPE_DEBUG) // debug BIOS
                {
                    bios_name = "dbg-bios.rom";
                }
                if (bios_name != NULL)
                    romdata->file = osd_fopen("neogeo", bios_name, OSD_FILETYPE_ROM, 0);
            }

            if (romdata->file != NULL)
            {
                disable_crc_check = 1;
            }
            else
            {
                romdata->file = osd_fopen(drv->name, name, OSD_FILETYPE_ROM, 0);
                options.bios_type = BIOS_TYPE_STANDARD;
            }
        }
        else
            romdata->file = osd_fopen(drv->name, name, OSD_FILETYPE_ROM, 0);
    }

    /* if that failed, attempt to open via CRC */
    sprintf(crc, "%08x", ROM_GETCRC(romp));
    for (drv = Machine->gamedrv; !romdata->file && drv; drv = drv->clone_of)
        romdata->file = osd_fopen(drv->name, crc, OSD_FILETYPE_ROM, 0);

    /* return the result */
    return (romdata->file != NULL);
}


/*-------------------------------------------------
    rom_fread - cheesy fread that fills with
    random data for a NULL file
-------------------------------------------------*/

static int rom_fread(struct rom_load_data *romdata, UINT8 *buffer, int length)
{
    /* files just pass through */
    if (romdata->file)
        return osd_fread(romdata->file, buffer, length);

    /* otherwise, fill with randomness */
    else
        fill_random(buffer, length);

    return length;
}


/*-------------------------------------------------
    read_rom_data - read ROM data for a single
    entry
-------------------------------------------------*/

static int read_rom_data(struct rom_load_data *romdata, const struct RomModule *romp)
{
    int datashift = ROM_GETBITSHIFT(romp);
    int datamask = ((1 << ROM_GETBITWIDTH(romp)) - 1) << datashift;
    int numbytes = ROM_GETLENGTH(romp);
    int groupsize = ROM_GETGROUPSIZE(romp);
    int skip = ROM_GETSKIPCOUNT(romp);
    int reversed = ROM_ISREVERSED(romp);
    int numgroups = (numbytes + groupsize - 1) / groupsize;
    UINT8 *base = romdata->regionbase + ROM_GETOFFSET(romp);
    int i;

    debugload("Loading ROM data: offs=%X len=%X mask=%02X group=%d skip=%d reverse=%d\n", ROM_GETOFFSET(romp), numbytes, datamask, groupsize, skip, reversed);

    /* make sure the length was an even multiple of the group size */
    if (numbytes % groupsize != 0)
    {
        printf("Error in RomModule definition: %s length not an even multiple of group size\n", ROM_GETNAME(romp));
        return -1;
    }

    /* make sure we only fill within the region space */
    if (ROM_GETOFFSET(romp) + numgroups * groupsize + (numgroups - 1) * skip > romdata->regionlength)
    {
        printf("Error in RomModule definition: %s out of memory region space\n", ROM_GETNAME(romp));
        return -1;
    }

    /* make sure the length was valid */
    if (numbytes == 0)
    {
        printf("Error in RomModule definition: %s has an invalid length\n", ROM_GETNAME(romp));
        return -1;
    }

    /* special case for simple loads */
    if (datamask == 0xff && (groupsize == 1 || !reversed) && skip == 0)
        return rom_fread(romdata, base, numbytes);

    /* chunky reads for complex loads */
    skip += groupsize;
    while (numbytes)
    {
        int evengroupcount = (sizeof(romdata->tempbuf) / groupsize) * groupsize;
        int bytesleft = (numbytes > evengroupcount) ? evengroupcount : numbytes;
        UINT8 *bufptr = romdata->tempbuf;

        /* read as much as we can */
        debugload("  Reading %X bytes into buffer\n", bytesleft);
        if (rom_fread(romdata, romdata->tempbuf, bytesleft) != bytesleft)
            return 0;
        numbytes -= bytesleft;

        debugload("  Copying to %08X\n", (UINT32)base);

        /* unmasked cases */
        if (datamask == 0xff)
        {
            /* non-grouped data */
            if (groupsize == 1)
                for (i = 0; i < bytesleft; i++, base += skip)
                    *base = *bufptr++;

            /* grouped data -- non-reversed case */
            else if (!reversed)
                while (bytesleft)
                {
                    for (i = 0; i < groupsize && bytesleft; i++, bytesleft--)
                        base[i] = *bufptr++;
                    base += skip;
                }

            /* grouped data -- reversed case */
            else
                while (bytesleft)
                {
                    for (i = groupsize - 1; i >= 0 && bytesleft; i--, bytesleft--)
                        base[i] = *bufptr++;
                    base += skip;
                }
        }

        /* masked cases */
        else
        {
            /* non-grouped data */
            if (groupsize == 1)
                for (i = 0; i < bytesleft; i++, base += skip)
                    *base = (*base & ~datamask) | ((*bufptr++ << datashift) & datamask);

            /* grouped data -- non-reversed case */
            else if (!reversed)
                while (bytesleft)
                {
                    for (i = 0; i < groupsize && bytesleft; i++, bytesleft--)
                        base[i] = (base[i] & ~datamask) | ((*bufptr++ << datashift) & datamask);
                    base += skip;
                }

            /* grouped data -- reversed case */
            else
                while (bytesleft)
                {
                    for (i = groupsize - 1; i >= 0 && bytesleft; i--, bytesleft--)
                        base[i] = (base[i] & ~datamask) | ((*bufptr++ << datashift) & datamask);
                    base += skip;
                }
        }
    }
    debugload("  All done\n");
    return ROM_GETLENGTH(romp);
}


/*-------------------------------------------------
    fill_rom_data - fill a region of ROM space
-------------------------------------------------*/

static int fill_rom_data(struct rom_load_data *romdata, const struct RomModule *romp)
{
    UINT32 numbytes = ROM_GETLENGTH(romp);
    UINT8 *base = romdata->regionbase + ROM_GETOFFSET(romp);

    /* make sure we fill within the region space */
    if (ROM_GETOFFSET(romp) + numbytes > romdata->regionlength)
    {
        printf("Error in RomModule definition: FILL out of memory region space\n");
        return 0;
    }

    /* make sure the length was valid */
    if (numbytes == 0)
    {
        printf("Error in RomModule definition: FILL has an invalid length\n");
        return 0;
    }

    /* fill the data */
    memset(base, ROM_GETCRC(romp) & 0xff, numbytes);
    return 1;
}


#if 0
/*-------------------------------------------------
    copy_rom_data - copy a region of ROM space
-------------------------------------------------*/

static int copy_rom_data(struct rom_load_data *romdata, const struct RomModule *romp)
{
    UINT8 *base = romdata->regionbase + ROM_GETOFFSET(romp);
    int srcregion = ROM_GETFLAGS(romp) >> 24;
    UINT32 numbytes = ROM_GETLENGTH(romp);
    UINT32 srcoffs = ROM_GETCRC(romp);
    UINT8 *srcbase;

    /* make sure we copy within the region space */
    if (ROM_GETOFFSET(romp) + numbytes > romdata->regionlength)
    {
        printf("Error in RomModule definition: COPY out of target memory region space\n");
        return 0;
    }

    /* make sure the length was valid */
    if (numbytes == 0)
    {
        printf("Error in RomModule definition: COPY has an invalid length\n");
        return 0;
    }

    /* make sure the source was valid */
    srcbase = memory_region(srcregion);
    if (!srcbase)
    {
        printf("Error in RomModule definition: COPY from an invalid region\n");
        return 0;
    }

    /* make sure we find within the region space */
    if (srcoffs + numbytes > memory_region_length(srcregion))
    {
        printf("Error in RomModule definition: COPY out of source memory region space\n");
        return 0;
    }

    /* fill the data */
    memcpy(base, srcbase + srcoffs, numbytes);
    return 1;
}
#endif


/*-------------------------------------------------
    process_rom_entries - process all ROM entries
    for a region
-------------------------------------------------*/

static int process_rom_entries(struct rom_load_data *romdata, const struct RomModule *romp)
{
    UINT32 lastflags = 0;

    /* loop until we hit the end of this region */
    while (!ROMENTRY_ISREGIONEND(romp))
    {
        /* if this is a continue entry, it's invalid */
        if (ROMENTRY_ISCONTINUE(romp))
        {
            printf("Error in RomModule definition: ROM_CONTINUE not preceded by ROM_LOAD\n");
            goto fatalerror;
        }

        /* if this is a reload entry, it's invalid */
        if (ROMENTRY_ISRELOAD(romp))
        {
            printf("Error in RomModule definition: ROM_RELOAD not preceded by ROM_LOAD\n");
            goto fatalerror;
        }

        /* handle fills */
        if (ROMENTRY_ISFILL(romp))
        {
            if (!fill_rom_data(romdata, romp++))
                goto fatalerror;
        }

#if 0
        /* handle copies */
        else if (ROMENTRY_ISCOPY(romp))
        {
            if (!copy_rom_data(romdata, romp++))
                goto fatalerror;
        }
#endif

        /* handle files */
        else if (ROMENTRY_ISFILE(romp))
        {
            const struct RomModule *baserom = romp;
            int explength = 0;

            /* open the file */
            debugload("Opening ROM file: %s\n", ROM_GETNAME(romp));
            if (!open_rom_file(romdata, romp))
                handle_missing_file(romdata, romp);

            /* loop until we run out of reloads */
            do
            {
                /* loop until we run out of continues */
                do
                {
                    struct RomModule modified_romp = *romp++;
                    int readresult;

                    /* handle flag inheritance */
                    if (!ROM_INHERITSFLAGS(&modified_romp))
                        lastflags = modified_romp._length & ROM_INHERITEDFLAGS;
                    else
                        modified_romp._length = (modified_romp._length & ~ROM_INHERITEDFLAGS) | lastflags;

					explength += UNCOMPACT_LENGTH(modified_romp._length);

                    /* attempt to read using the modified entry */
                    readresult = read_rom_data(romdata, &modified_romp);
                    if (readresult == -1)
                        goto fatalerror;
                }
                while (ROMENTRY_ISCONTINUE(romp));

                /* if this was the first use of this file, verify the length and CRC */
                if (baserom)
                {
                    debugload("Verifying length (%X) and CRC (%08X)\n", explength, ROM_GETCRC(baserom));
                    verify_length_and_crc(romdata, ROM_GETNAME(baserom), explength, ROM_GETCRC(baserom));
                    debugload("Verify succeeded\n");
                }

                /* reseek to the start and clear the baserom so we don't reverify */
                if (romdata->file)
                    osd_fseek(romdata->file, 0, SEEK_SET);
                baserom = NULL;
				explength = 0;
            }
            while (ROMENTRY_ISRELOAD(romp));

            /* close the file */
            if (romdata->file)
            {
                debugload("Closing ROM file\n");
                osd_fclose(romdata->file);
                romdata->file = NULL;
            }
        }
    }
    return 1;

    /* error case */
fatalerror:
    if (romdata->file)
        osd_fclose(romdata->file);
    romdata->file = NULL;
    return 0;
}


/*-------------------------------------------------
    rom_load_new - new, more flexible ROM
    loading system
-------------------------------------------------*/

int rom_load_new(const struct RomModule *romp)
{
    const struct RomModule *regionlist[REGION_MAX];
    const struct RomModule *region;
    static struct rom_load_data romdata;
    int regnum;

    if (!options.disable_patch)
    {
        if (!strcmp(Machine->gamedrv->name, "fightfev") ||
			!strcmp(Machine->gamedrv->name, "garoup"))
            disable_crc_check = 1;
    }
    else
        disable_crc_check = 0;

    /* reset the region list */
    memset((char *)regionlist, 0, sizeof(regionlist));

    /* reset the romdata struct */
    memset(&romdata, 0, sizeof(romdata));
    romdata.romstotal = count_roms(romp);

    /* loop until we hit the end */
    for (region = romp, regnum = 0; region; region = rom_next_region(region), regnum++)
    {
        int regiontype = ROMREGION_GETTYPE(region);

        debugload("Processing region %02X (length=%X)\n", regiontype, ROMREGION_GETLENGTH(region));

        /* the first entry must be a region */
        if (!ROMENTRY_ISREGION(region))
        {
            printf("Error: missing ROM_REGION header\n");
            return 1;
        }

        /* if sound is disabled and it's a sound-only region, skip it */
        if (Machine->sample_rate == 0 && ROMREGION_ISSOUNDONLY(region))
            continue;

        /* allocate memory for the region */
        if (new_memory_region(regiontype, ROMREGION_GETLENGTH(region), ROMREGION_GETFLAGS(region)))
        {
            printf("Error: unable to allocate memory for region %d\n", regiontype);
            return 1;
        }

        /* remember the base and length */
        romdata.regionlength = memory_region_length(regiontype);
        romdata.regionbase = memory_region(regiontype);
        debugload("Allocated %X bytes @ %08X\n", romdata.regionlength, (UINT32)romdata.regionbase);

        /* clear the region if it's requested */
        if (ROMREGION_ISERASE(region))
            memset(romdata.regionbase, ROMREGION_GETERASEVAL(region), romdata.regionlength);

        /* or if it's sufficiently small (<= 4MB) */
        else if (romdata.regionlength <= 0x400000)
            memset(romdata.regionbase, 0, romdata.regionlength);

#ifdef MAME_DEBUG
        /* if we're debugging, fill region with random data to catch errors */
        else
            fill_random(romdata.regionbase, romdata.regionlength);
#endif

        /* now process the entries in the region */
        if (!process_rom_entries(&romdata, region + 1))
            return 1;

        /* add this region to the list */
        if (regiontype < REGION_MAX)
            regionlist[regiontype] = region;
    }

    /* post-process the regions */
    for (regnum = 0; regnum < REGION_MAX; regnum++)
    {
        if (regionlist[regnum])
        {
            debugload("Post-processing region %02X\n", regnum);
            romdata.regionlength = memory_region_length(regnum);
            romdata.regionbase = memory_region(regnum);
            region_post_process(&romdata, regionlist[regnum]);
        }
    }

    /* display the results and exit */
    return display_rom_load_results(&romdata);
}
