#include "driver.h"
#include <ctype.h>
#include "ui_text.h" /* LBO 042400 */
#include "drivers/neogeo.h"
#ifdef KAILLERA
#include "KailleraChat.h"
#endif

static struct RunningMachine machine;
struct RunningMachine *Machine = &machine;
static const struct GameDriver *gamedrv;
static const struct MachineDriver *drv;
static struct osd_bitmap *real_scrbitmap;

/* Variables to hold the status of various game options */
struct GameOptions	options;

void *record;   /* for -record */
void *playback; /* for -playback */
#ifdef MAME_DEBUG
int mame_debug; /* !0 when -debug option is specified */
#endif

int bailing;	/* set to 1 if the startup is aborted to prevent multiple error messages */

static int settingsloaded;

static int leds_status;


int init_machine(void);
void shutdown_machine(void);
int run_machine(void);

int run_game(int game)
{
	int err;

	/* copy some settings into easier-to-handle variables */
	record     = options.record;
	playback   = options.playback;
#ifdef MAME_DEBUG
	mame_debug = options.mame_debug;
#endif

	neogeo_change_driver(game);

	Machine->gamedrv = gamedrv = drivers[game];
	Machine->drv = drv = gamedrv->drv;

	keyname_init();
	portname_init();

	if (custom_button_init())
	{
		int i;

		i = read_combo_ex();

		if (i != 0)
			logerror("Combo loaded: %d\n", i);
	}

	/* copy configuration */
	Machine->sample_rate = options.samplerate;

	/* Do the work*/
	err = 1;
	bailing = 0;

	if (osd_init() == 0)
	{
		if (init_machine() == 0)
		{
			if (run_machine() == 0)
				err = 0;
			else if (!bailing)
			{
				bailing = 1;
				logerror("Unable to start machine emulation\n");
			}

			shutdown_machine();
		}
		else if (!bailing)
		{
			bailing = 1;
			logerror("Unable to initialize machine emulation\n");
		}

		osd_exit();
	}
	else if (!bailing)
	{
		bailing = 1;
		logerror("Unable to initialize system\n");
	}

	custom_button_end();

	return err;
}



/***************************************************************************

  Initialize the emulated machine (load the roms, initialize the various
  subsystems...). Returns 0 if successful.

***************************************************************************/
int init_machine(void)
{
	if (code_init() != 0)
	{
		logerror("code_init failed\n");
		goto out;
	}

	memset(&Machine->memory_region, 0, sizeof(Machine->memory_region));

	if (gamedrv->input_ports)
	{
		Machine->input_ports = input_port_allocate(gamedrv->input_ports);
		if (!Machine->input_ports)
		{
			logerror("could not allocate Machine->input_ports\n");
			goto out_code;
		}
		Machine->input_ports_default = input_port_allocate(gamedrv->input_ports);
		if (!Machine->input_ports_default)
		{
			logerror("could not allocate Machine->input_ports_default\n");
			input_port_free(Machine->input_ports);
			Machine->input_ports = 0;
			goto out_code;
		}
	}

	if (readroms() != 0)
	{
		logerror("readroms failed\n");
		goto out_free;
	}

	if (!options.disable_patch)
	{
		if (strcmp(Machine->gamedrv->name, "garoup") == 0)
			neogeo_garou_patch();
		else if (strcmp(Machine->gamedrv->name, "fightfev") == 0)
			neogeo_fightfev_patch();
	}

	/* first of all initialize the memory handlers, which could be used by the */
	/* other initialization routines */
	cpu_init();

	/* load input ports settings (keys, dip switches, and so on) */
	settingsloaded = load_input_port_settings();

	if( !memory_init() )
	{
		logerror("memory_init failed\n");
		goto out_free;
	}

	if (gamedrv->driver_init) (*gamedrv->driver_init)();

	return 0;

out_free:
	input_port_free(Machine->input_ports);
	Machine->input_ports = 0;
	input_port_free(Machine->input_ports_default);
	Machine->input_ports_default = 0;
out_code:
	code_close();
out:
	return 1;
}



void shutdown_machine(void)
{
	int i;


    /* ASG 971007 free memory element map */
	memory_shutdown();

	/* free the memory allocated for ROM and RAM */
	for (i = 0;i < MAX_MEMORY_REGIONS;i++)
		free_memory_region(i);

	/* free the memory allocated for input ports definition */
	input_port_free(Machine->input_ports);
	Machine->input_ports = 0;
	input_port_free(Machine->input_ports_default);
	Machine->input_ports_default = 0;

	code_close();
}



static void vh_close(void)
{
	int i;


	for (i = 0;i < MAX_GFX_ELEMENTS;i++)
	{
		freegfx(Machine->gfx[i]);
		Machine->gfx[i] = 0;
	}
	uifont_freefont();
#ifdef MAME_DEBUG
	if (Machine->debugger_font)
	{
		freegfx(Machine->debugger_font);
		Machine->debugger_font = NULL;
	}
#endif
	osd_close_display();
	if (Machine->scrbitmap)
	{
		bitmap_free(Machine->scrbitmap);
		Machine->scrbitmap = NULL;
	}
#ifdef MAME_DEBUG
	if (Machine->debug_bitmap)
	{
		osd_free_bitmap(Machine->debug_bitmap);
		Machine->debug_bitmap = NULL;
	}
#endif

	palette_stop();
}



static int vh_open(void)
{
	int i;
	int width,height;

	for (i = 0;i < MAX_GFX_ELEMENTS;i++) Machine->gfx[i] = 0;
	Machine->uifont = NULL;
	Machine->uifont2 = NULL;
#ifdef MAME_DEBUG
	Machine->debugger_font = NULL;
#endif

	if (palette_start() != 0)
	{
		vh_close();
		return 1;
	}


	/* convert the gfx ROMs into character sets. This is done BEFORE calling the driver's */
	/* convert_color_prom() routine (in palette_init()) because it might need to check the */
	/* Machine->gfx[] data */
	if (drv->gfxdecodeinfo)
	{
		for (i = 0;i < MAX_GFX_ELEMENTS && drv->gfxdecodeinfo[i].memory_region != -1;i++)
		{
			int reglen = 8*memory_region_length(drv->gfxdecodeinfo[i].memory_region);
			struct GfxLayout glcopy;
			int j;


			memcpy(&glcopy,drv->gfxdecodeinfo[i].gfxlayout,sizeof(glcopy));

			if (IS_FRAC(glcopy.total))
				glcopy.total = reglen / glcopy.charincrement * FRAC_NUM(glcopy.total) / FRAC_DEN(glcopy.total);
			for (j = 0;j < MAX_GFX_PLANES;j++)
			{
				if (IS_FRAC(glcopy.planeoffset[j]))
				{
					glcopy.planeoffset[j] = FRAC_OFFSET(glcopy.planeoffset[j]) +
							reglen * FRAC_NUM(glcopy.planeoffset[j]) / FRAC_DEN(glcopy.planeoffset[j]);
				}
			}
			for (j = 0;j < MAX_GFX_SIZE;j++)
			{
				if (IS_FRAC(glcopy.xoffset[j]))
				{
					glcopy.xoffset[j] = FRAC_OFFSET(glcopy.xoffset[j]) +
							reglen * FRAC_NUM(glcopy.xoffset[j]) / FRAC_DEN(glcopy.xoffset[j]);
				}
				if (IS_FRAC(glcopy.yoffset[j]))
				{
					glcopy.yoffset[j] = FRAC_OFFSET(glcopy.yoffset[j]) +
							reglen * FRAC_NUM(glcopy.yoffset[j]) / FRAC_DEN(glcopy.yoffset[j]);
				}
			}

			if ((Machine->gfx[i] = decodegfx(memory_region(drv->gfxdecodeinfo[i].memory_region)
					+ drv->gfxdecodeinfo[i].start,
					&glcopy)) == 0)
			{
				vh_close();

				bailing = 1;
				logerror("Out of memory decoding gfx\n");

				return 1;
			}
			Machine->gfx[i]->colortable = &Machine->pens[drv->gfxdecodeinfo[i].color_codes_start];
			Machine->gfx[i]->total_colors = drv->gfxdecodeinfo[i].total_color_codes;
		}
	}

	width = drv->screen_width;
	height = drv->screen_height;

	Machine->scrbitmap = bitmap_alloc(width, height);
	if (!Machine->scrbitmap)
	{
		vh_close();
		return 1;
	}

#ifdef MAME_DEBUG
	if (mame_debug)
	{
		Machine->debug_bitmap = bitmap_alloc(options.debug_width,options.debug_height);
		if (!Machine->debug_bitmap)
		{
			vh_close();
			return 1;
		}
	}
#endif

	width = drv->default_visible_area.max_x - drv->default_visible_area.min_x + 1;
	height = drv->default_visible_area.max_y - drv->default_visible_area.min_y + 1;

	/* create the display bitmap, and allocate the palette */
	if (osd_create_display(width,height,drv->frames_per_second))
	{
		vh_close();
		return 1;
	}

	set_visible_area(
			drv->default_visible_area.min_x,
			drv->default_visible_area.max_x,
			drv->default_visible_area.min_y,
			drv->default_visible_area.max_y);

	/* build our private user interface font */
	/* This must be done AFTER osd_create_display() so the function knows the */
	/* resolution we are running at and can pick a different font depending on it. */
	/* It must be done BEFORE palette_init() because that will also initialize */
	/* (through osd_allocate_colors()) the uifont colortable. */
	if (!uifont_buildfont())
	{
		vh_close();
		return 1;
	}
#ifdef MAME_DEBUG
	if (mame_debug)
	{
		if (NULL == (Machine->debugger_font = build_debugger_font()))
		{
			vh_close();
			return 1;
		}
	}
#endif

	/* initialize the palette - must be done after osd_create_display() */
	if (palette_init())
	{
		vh_close();
		return 1;
	}

	leds_status = 0;

	return 0;
}



/***************************************************************************

  This function takes care of refreshing the screen, processing user input,
  and throttling the emulation speed to obtain the required frames per second.

***************************************************************************/

int updatescreen(void)
{
	/* update sound */
	sound_update();

	if (osd_skip_this_frame() == 0)
	{
		profiler_mark(PROFILER_VIDEO);
		draw_screen();	/* update screen */
		profiler_mark(PROFILER_END);
	}

	/* the user interface must be called between vh_update() and osd_update_video_and_audio(), */
	/* to allow it to overlay things on the game display. We must call it even */
	/* if the frame is skipped, to keep a consistent timing. */
	if (handle_user_interface(real_scrbitmap))
		/* quit if the user asked to */
		return 1;

	update_video_and_audio();

	return 0;
}


/***************************************************************************

  Draw screen with overlays and backdrops (not yet)

***************************************************************************/

static int bitmap_dirty;

void draw_screen(void)
{
	(*Machine->drv->vh_update)(Machine->scrbitmap,bitmap_dirty);  /* update screen */

	bitmap_dirty = 0;
}

void schedule_full_refresh(void)
{
	bitmap_dirty = 1;
}


/***************************************************************************

  Calls OSD layer handling overlays and backdrops (not yet)

***************************************************************************/
void update_video_and_audio(void)
{
#ifdef MAME_DEBUG
	debug_trace_delay = 0;
#endif
	osd_update_video_and_audio(real_scrbitmap,
#ifdef MAME_DEBUG
	                           Machine->debug_bitmap,
#endif
	                           leds_status);
}


/***************************************************************************

  Run the emulation. Start the various subsystems and the CPU emulation.
  Returns non zero in case of error.

***************************************************************************/
int run_machine(void)
{
#ifdef KAILLERA
	extern int kPlay;
#endif
	int res = 1;


	if (vh_open() == 0)
	{
		if (drv->vh_start == 0 || (*drv->vh_start)() == 0)      /* start the video hardware */
		{
			if (sound_start() == 0) /* start the audio hardware */
			{
				int	region;

#ifdef KAILLERA
				if (kPlay)
					KailleraChatInit();
#endif
				
//				real_scrbitmap = artwork_overlay ? overlay_real_scrbitmap : Machine->scrbitmap;
				real_scrbitmap = Machine->scrbitmap;

				/* free memory regions allocated with REGIONFLAG_DISPOSE (typically gfx roms) */
				for (region = 0; region < MAX_MEMORY_REGIONS; region++)
				{
					if (Machine->memory_region[region].flags & ROMREGION_DISPOSE)
					{
						int i;

						/* invalidate contents to avoid subtle bugs */
						for (i = 0;i < memory_region_length(region);i++)
							memory_region(region)[i] = rand();
						MyFree(Machine->memory_region[region].base);
						Machine->memory_region[region].base = 0;
					}
				}

				if (settingsloaded == 0)
				{
					/* if there is no saved config, it must be first time we run this game, */
					/* so show the disclaimer. */
#ifdef KAILLERA
					if (!kPlay)	/* disable startup screen with network game*/
#endif
					if (showcopyright(real_scrbitmap)) goto userquit;
				}

				if (showgamewarnings(real_scrbitmap) == 0)  /* show info about incorrect behaviour (wrong colors etc.) */
				{
					init_user_interface();

					/* disable cheat if no roms */
					if (!gamedrv->rom) options.cheat = 0;

					if (options.cheat) InitCheat();

					if (drv->nvram_handler)
					{
						void *f;
#ifdef KAILLERA
					if (kPlay)
					{
						(*drv->nvram_handler)(0,0);
					}
					else
					{
#endif
						f = osd_fopen(Machine->gamedrv->name,0,OSD_FILETYPE_NVRAM,0);
						(*drv->nvram_handler)(f,0);
						if (f) osd_fclose(f);
					}
#ifdef KAILLERA
					}
#endif
					cpu_run();      /* run the emulation! */

#ifdef KAILLERA
	if (!kPlay)
	{
#endif
					if (drv->nvram_handler)
					{
						void *f;

						if ((f = osd_fopen(Machine->gamedrv->name,0,OSD_FILETYPE_NVRAM,1)) != 0)
						{
							(*drv->nvram_handler)(f,1);
							osd_fclose(f);
						}
					}
#ifdef KAILLERA
 	}
#endif

					if (options.cheat) StopCheat();

					/* save input ports settings */
					save_input_port_settings();
				}

userquit:
				/* the following MUST be done after hiscore_save() otherwise */
				/* some 68000 games will not work */
				sound_stop();
				if (drv->vh_stop) (*drv->vh_stop)();

#ifdef KAILLERA
				if (kPlay)
					KailleraChatExit();
#endif
				res = 0;
			}
			else if (!bailing)
			{
				bailing = 1;
				logerror("Unable to start audio emulation\n");
			}
		}
		else if (!bailing)
		{
			bailing = 1;
			logerror("Unable to start video emulation\n");
		}

		vh_close();
	}
	else if (!bailing)
	{
		bailing = 1;
		logerror("Unable to start video emulation\n");
	}

	return res;
}



int mame_highscore_enabled(void)
{
	/* disable high score when record/playback is on */
	if (record != 0 || playback != 0) return 0;

	/* disable high score when cheats are used */
	if (he_did_cheat != 0) return 0;

	return 1;
}


void set_led_status(int num,int on)
{
	if (on) leds_status |=	(1 << num);
	else	leds_status &= ~(1 << num);
}

