/***************************************************************************

  memmanager.c

  Memory allocate handling for NMAME X II

***************************************************************************/

#ifdef MEM_DEBUG
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "memmngr.h"


#define LOG_ERROR          1
#define MAX_MEMORY_ENTRIES 512

static struct my_memory_t {
	void   *ptr;
	size_t size;
	char   *file;
	int    line;
	int    in_emulation;
} memory_entries[MAX_MEMORY_ENTRIES];


static int total_leaked  = 0;
static int max_count     = 0;
static int current_count = 0;

extern int in_emulation;

static void CLIB_DECL memory_manager_exit(void);



void memory_manager_init(void)
{
	int i;

	for (i = 0; i < MAX_MEMORY_ENTRIES; i++)
		memset(&memory_entries[i], 0, sizeof(struct my_memory_t));

	in_emulation = 0;

	atexit(memory_manager_exit);
}


void CLIB_DECL memory_manager_exit(void)
{
	int i, leaked_count = 0;
#if LOG_ERROR
	FILE *fp;

	remove("memdebug.log");
	fp = fopen("memdebug.log", "w");
	if (fp)
		fprintf(fp, "[NMAME X II - leaked memory list]\n\n");
#endif

	for (i = 0; i < MAX_MEMORY_ENTRIES; i++)
	{
		if (memory_entries[i].ptr)
		{
#if LOG_ERROR
			if (fp)
			{
				fprintf(fp, "[%3d - %s] : %i %8d bytes (in %s - line %d)\n",
						i,
						(memory_entries[i].in_emulation) ? "EMU" : "GUI",
						memory_entries[i].ptr,
						memory_entries[i].size,
						memory_entries[i].file,
						memory_entries[i].line);
			}
#endif
			total_leaked += memory_entries[i].size;
			leaked_count++;
			free(memory_entries[i].ptr);
			memory_entries[i].ptr  = NULL;
		}
	}

#if LOG_ERROR
	if (fp)
	{
		fprintf(fp, "\nmax allocate count: %d\n", max_count);
		fprintf(fp, "total leaked: %d (%d bytes)\n", leaked_count, total_leaked);
		fclose(fp);

		if (total_leaked == 0)
			remove("memdebug.log");
	}
#endif
}


void *my_malloc(size_t size, char *file, int line)
{
	int i, found = 0;

	for (i = 0; i < MAX_MEMORY_ENTRIES; i++)
	{
		if (memory_entries[i].ptr == NULL)
		{
			found = 1;
			break;
		}
	}

	if (found)
	{
		memory_entries[i].ptr = malloc(size);
		if (memory_entries[i].ptr)
		{
			memset(memory_entries[i].ptr, 0, size);
			memory_entries[i].size = size;
			memory_entries[i].file = file;
			memory_entries[i].line = line;
			memory_entries[i].in_emulation = in_emulation;

			current_count++;
			if (max_count < current_count)
				max_count = current_count;
		}
		return memory_entries[i].ptr;
	}
	return NULL;
}


void *my_calloc(size_t n, size_t size, char *file, int line)
{
	int i, found = 0;

	for (i = 0; i < MAX_MEMORY_ENTRIES; i++)
	{
		if (memory_entries[i].ptr == NULL)
		{
			found = 1;
			break;
		}
	}

	if (found)
	{
		memory_entries[i].ptr = calloc(n, size);
		if (memory_entries[i].ptr)
		{
			memset(memory_entries[i].ptr, 0, n * size);
			memory_entries[i].size = n * size;
			memory_entries[i].file = file;
			memory_entries[i].line = line;
			memory_entries[i].in_emulation = in_emulation;

			current_count++;
			if (max_count < current_count)
				max_count = current_count;
		}
		return memory_entries[i].ptr;
	}

	return NULL;
}


void *my_realloc(void *ptr, size_t size, char *file, int line)
{
	int i, found = 0;
	void *new_ptr;

	if (ptr == NULL)
		return my_malloc(size, file, line);

	for (i = 0; i < MAX_MEMORY_ENTRIES; i++)
	{
		if (memory_entries[i].ptr == ptr)
		{
			found = 1;
			break;
		}
	}

	if (found)
	{
		new_ptr = realloc(ptr, size);
		if (new_ptr)
		{
			if (new_ptr != ptr)
			{
				memory_entries[i].ptr  = new_ptr;
				memory_entries[i].size = size;
				memory_entries[i].file = file;
				memory_entries[i].line = line;
				memory_entries[i].in_emulation = in_emulation;
			}

			return new_ptr;
		}

		my_free(memory_entries[i].ptr);
	}

	return NULL;
}


void my_free(void *ptr)
{
	int i, found = 0;

	if (ptr == NULL)
		return;

	for (i = 0; i < MAX_MEMORY_ENTRIES; i++)
	{
		if (memory_entries[i].ptr == ptr)
		{
			found = 1;
			break;
		}
	}

	if (found)
	{
		current_count--;
		free(memory_entries[i].ptr);
		memset(&memory_entries[i], 0, sizeof(struct my_memory_t));
	}
}
#endif
